# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging

import numpy as np

from magnopy.energy.checkers import (
    _check_cone_axis,
    _check_directions,
    _check_spiral_vector,
)
from magnopy.energy.optimize import _bfgs as _optimize
from magnopy.spinham.hamiltonian import SpinHamiltonian
from magnopy.units.inside import ENERGY
from magnopy.units.si import BOHR_MAGNETON

_logger = logging.getLogger(__name__)

__all__ = ["C1", "C2", "C5"]  # , "C3", "C4"]

# Convert to the internal units of energy
BOHR_MAGNETON = BOHR_MAGNETON / ENERGY


class _Energy:
    r"""
    Generic energy class.
    """

    def __init__(self, spinham: SpinHamiltonian):
        self._magnetic_field = None
        # Parameters of the Hamiltonian, private
        self._atom_indices = dict(
            [(atom, i) for i, atom in enumerate(spinham.magnetic_atoms)]
        )
        self._g_factors = [atom.g_factor for atom in spinham.magnetic_atoms]
        self._spins = [atom.spin for atom in spinham.magnetic_atoms]

        self._I = len(self._spins)

        self._cell = spinham.cell

        # Force double counting notation
        previous_dc = spinham.double_counting
        previous_factor = spinham.exchange_factor
        previous_spin_normalized = spinham.spin_normalized
        spinham.double_counting = True
        spinham.exchange_factor = 0.5
        spinham.spin_normalized = True

        # Get all bonds of the Hamiltonian.
        self._bonds = []
        for atom1, atom2, R, J in spinham.exchange_like:
            i = self._atom_indices[atom1]
            j = self._atom_indices[atom2]
            d = R @ spinham.cell
            self._bonds.append([i, j, d, J])

        # Return original notation of SpinHamiltonian
        spinham.double_counting = previous_dc
        spinham.exchange_factor = previous_factor
        spinham.spin_normalized = previous_spin_normalized

    @property
    def magnetic_field(self):
        r"""
        Magnetic field vector.
        """
        return self._magnetic_field

    @magnetic_field.setter
    def magnetic_field(self, new_value):
        # Check that the input is array-like and convertible to floats
        try:
            new_value = np.array(new_value, dtype=float)
        except:
            raise ValueError(f"Magnetic field is not array-like: {new_value}.")

        # Check that the shape and size are correct
        if new_value.shape != (3,):
            raise ValueError(
                f"Magnetic field has to have shape (3,), got {new_value.shape}."
            )

        self._magnetic_field = new_value

    @property
    def spiral_vector(self):
        raise RuntimeError(
            f"Discrete spiral vector is not defined for the energy type "
            f"{self.__class__.__name__}."
        )

    @spiral_vector.setter
    def spiral_vector(self, new_value):
        raise RuntimeError(
            f"Discrete spiral vector is not defined for the energy type "
            f"{self.__class__.__name__}."
        )

    def energy(*args, **kwargs):
        raise NotImplementedError

    def gradient(*args, **kwargs):
        raise NotImplementedError

    def _F(*args, **kwargs):
        raise NotImplementedError

    def _grad_F(*args, **kwargs):
        raise NotImplementedError

    def _to_x(true_variables):
        raise NotImplementedError

    def _update(true_variables, x):
        raise NotImplementedError

    def _update_difference(func_k, func_kp1, grad_kp1):
        raise NotImplementedError

    def optimize(*args, **kwargs):
        raise NotImplementedError


class C1(_Energy):
    def __init__(self, spinham: SpinHamiltonian):
        super().__init__(spinham)

        self.J = np.zeros((self._I, self._I, 3, 3), dtype=float)

        for i, j, _, J in self._bonds:
            self.J[i, j] += J.matrix

    def energy(self, directions, check_input=True) -> float:
        r"""
        Compute the energy of the system with respect to the provided directions of
        spins. Computed according to the formula FIXME of FIXME.

        Parameters
        ----------
        directions : (I, 3) or (3,) |array-like|_
            Orientation of the spin vectors.
            If ``I = 1``, then both ``(1,3)`` and ``(3,)`` shaped inputs are accepted.
            The vectors are normalized to one, i.e. only the direction of the vectors
            is important.
        check_input : bool, default=True
            Whether to check the correctness of the input. Better keep it ``True``
            always, unless you need to compute energy thousands of times and you control
            the input correctness externally.

        Returns
        -------
        energy : float
            Energy of the system.
        """

        if check_input:
            directions = _check_directions(self._I, directions)

        energy = 0

        # Compute exchange and single-ion-like anisotropy energy
        energy += 0.5 * np.einsum(
            "i,j,ix,jy,ijxy",
            self._spins,
            self._spins,
            directions,
            directions,
            self.J,
        )

        # Compute zeeman energy
        if self.magnetic_field is not None:
            energy += BOHR_MAGNETON * np.einsum(
                "i,i,j,ij",
                self._g_factors,
                self._spins,
                self.magnetic_field,
                directions,
            )

        return energy

    def gradient(self, directions, check_input=True) -> np.ndarray:
        r"""
        Compute the gradient of the energy with respect to the provided directions of
        spins. Computed according to the formula FIXME of FIXME.

        Parameters
        ----------
        directions : (I, 3) or (3,) |array-like|_
            Orientation of the spin vectors.
            If ``I = 1``, then both ``(1,3)`` and ``(3,)`` shaped inputs are accepted.
            The vectors are normalized to one, i.e. only the direction of the vectors
            is important.
        check_input : bool, default=True
            Whether to check the correctness of the input. Better keep it ``True``
            always, unless you need to compute energy thousands of times and you control
            the input correctness externally.

        Returns
        -------
        gradient : (I,3) :numpy:`ndarray`
            Gradient of the energy. Form:

            .. code-block:: python

                [
                    [ dE/de1x, dE/de1y, dE/de1z ],
                    [ dE/de2x, dE/de2y, dE/de2z ],
                    ...
                    [ dE/deIx, dE/deIy, dE/deIz ]
                ]
        """

        if check_input:
            directions = _check_directions(self._I, directions)

        gradient = np.zeros_like(directions, dtype=float)

        # Derivative of the Zeeman term
        if self.magnetic_field is not None:
            gradient += BOHR_MAGNETON * np.einsum(
                "b,b,x->bx",
                self._g_factors,
                self._spins,
                self.magnetic_field,
            )

        # Derivative of bilinear term
        gradient += np.einsum(
            "ak,a,b,abkx->bx",
            directions,
            self._spins,
            self._spins,
            self.J,
        )

        return gradient

    def _F(self, true_variables) -> float:
        # Remove input check as this function is specific for the minimization procedure
        return self.energy(true_variables, check_input=False)

    def _grad_F(self, true_variables) -> np.ndarray:
        r"""
        Compute the gradient of the energy as a function of
        :math:`(a^x_1, a^y_1, a^z_1, ..., a^x_I, a^y_I, a^z_I)` for the C1 sub-type.

        Parameters
        ----------
        true_variables : (I, 3) :numpy:`ndarray`
            Directions of the spin vectors.

        Returns
        -------
        gradient : (I*3,) :numpy:`ndarray`
            Gradient of the energy. Form:

            .. code-block:: python

                [ dF/da1x, dF/da1y, dF/da1z, ..., dF/daIx, dF/daIy, dF/daIz ]
        """

        directions = true_variables

        # Remove input check as this function is specific for the minimization procedure
        gradient = self.gradient(directions, check_input=False)

        torque = np.cross(directions, gradient)

        return torque.flatten()

    def _to_x(self, true_variables):
        r"""

        Parameters
        ----------
        true_variables : (I, 3) :numpy:`ndarray`
            Original direction of the spin vectors.

        Returns
        -------
        x : (I*3,) :numpy:`ndarray`
            Direction of the spin vectors parameterized with the skew-symmetric matrix.
        """

        return np.zeros(true_variables.size, dtype=float)

    def _update(self, true_variables, x):
        r"""

        Parameters
        ----------
        true_variables : (I, 3) :numpy:`ndarray`
            Original direction of the spin vectors.
        x : (I*3,) :numpy:`ndarray`
            Direction of the spin vectors parameterized with the skew-symmetric matrix.

        Returns
        -------
        directions : (I, 3) :numpy:`ndarray`
            Rotated set of direction vectors.
        """

        directions = true_variables.copy()

        ax = x[::3]
        ay = x[1::3]
        az = x[2::3]

        thetas = np.sqrt(ax**2 + ay**2 + az**2)

        r = []
        for i in range(len(thetas)):
            theta = thetas[i]

            if theta < np.finfo(float).eps:
                continue

            r = np.array([ax[i], ay[i], az[i]]) / theta

            directions[i] = (
                np.cos(theta) * directions[i]
                + np.sin(theta) * np.cross(r, directions[i])
                + (1 - np.cos(theta)) * r * (r @ directions[i])
            )

        return directions

    def _update_difference(self, func_k, func_kp1, grad_kp1):
        r"""
        Computes the difference between two consecutive steps of the optimization.

        Parameters
        ----------
        func_k : float
            Energy at the current step.
        func_kp1 : float
            Energy at the next step.
        grad_kp1 : (I*3,) :numpy:`ndarray`
            Gradient of the energy at the next state.

        Returns
        -------
        difference : (2,) :numpy:`ndarray`
            Difference between the two consecutive steps of the optimization.
        """

        return np.array(
            [
                abs(func_kp1 - func_k),
                np.linalg.norm(grad_kp1.reshape(grad_kp1.size // 3, 3), axis=1).max(),
            ]
        )

    def _hessian(self, true_variables):
        h = 1e-6

        x_0 = self._to_x(true_variables)

        hessian = np.zeros((x_0.size, x_0.size), dtype=float)

        for i in range(x_0.size):
            for j in range(x_0.size):
                x_pp = x_0.copy()
                x_pm = x_0.copy()
                x_mp = x_0.copy()
                x_mm = x_0.copy()

                x_pp[i] += h
                x_pp[j] += h

                x_pm[i] += h
                x_pm[j] -= h

                x_mp[i] -= h
                x_mp[j] += h

                x_mm[i] -= h
                x_mm[j] -= h

                tv_pp = self._update(true_variables, x_pp)
                tv_pm = self._update(true_variables, x_pm)
                tv_mp = self._update(true_variables, x_mp)
                tv_mm = self._update(true_variables, x_mm)

                hessian[i, j] = (
                    (self._F(tv_pp) - self._F(tv_pm) - self._F(tv_mp) + self._F(tv_mm))
                    / 4
                    / h
                    / h
                )

        return hessian

    def optimize(
        self, directions_ig=None, energy_tolerance=1e-5, torque_tolerance=1e-5
    ):
        r"""
        Optimize the energy with respect to the directions of spins in the unit cell.

        Parameters
        ----------
        directions_ig : (I, 3) or (3,) |array-like|_, optional
            Initial guess for the direction of the spin vectors.
        energy_tolerance : float, default 1e-5
            Energy tolerance for the two consecutive steps of the optimization.
        torque_tolerance : float, default 1e-5
            Torque tolerance for the two consecutive steps of the optimization.

        Returns
        -------
        optimized_directions : (I, 3) :numpy:`ndarray`
            Optimized direction of the spin vectors.
        """

        if directions_ig is None:
            directions_ig = np.random.uniform(low=-1, high=1, size=(self._I, 3))
            _logger.debug("Initial guess for the optimization is random (C1).")

        # Check the input correctness and normalize vectors.
        directions = _check_directions(self._I, directions_ig)

        hessian = self._hessian(directions)

        print(hessian)

        optimized_directions = _optimize(
            self,
            true_variables_0=directions,
            hessian_0=hessian,
            tolerance=np.array([energy_tolerance, torque_tolerance]),
            func=self._F,
            grad=self._grad_F,
            update=self._update,
            to_x=self._to_x,
            update_difference=self._update_difference,
        )

        return optimized_directions


class C2(_Energy):
    def __init__(self, spinham: SpinHamiltonian):
        super().__init__(spinham)

        self._spiral_vector = None
        self.spiral_vector = np.ones(3, dtype=float) / 2

    @property
    def spiral_vector(self):
        return self._spiral_vector

    @spiral_vector.setter
    def spiral_vector(self, new_value):
        # Check that the input is array-like and convertible to floats
        try:
            new_value = np.array(new_value, dtype=float)
        except:
            raise ValueError(f"spiral_vector is not array-like: {new_value}.")

        # Check that the shape and size are correct
        if new_value.shape != (3,):
            raise ValueError(
                f"spiral_vector has to have shape (3,), got {new_value.shape}."
            )

        self._spiral_vector = new_value

    def energy(self, directions, cone_axis, check_input=True) -> float:
        r"""
        Compute the energy of the system with respect to the provided directions of
        spins, cone axis and spiral vector. Computed according to the formula FIXME of FIXME.

        Parameters
        ----------
        directions : (I, 3) or (3,) |array-like|_
            Orientation of the spin vectors.
            If ``I = 1``, then both ``(1,3)`` and ``(3,)`` shaped inputs are accepted.
            The vectors are normalized to one, i.e. only the direction of the vectors
            is important.
        cone_axis : (3,) |array-like|_
            Cone axis of the spiral cone state.
        check_input : bool, default=True
            Whether to check the correctness of the input. Better keep it ``True``
            always, unless you need to compute energy thousands of times and you control
            the input correctness externally.

        Returns
        -------
        energy : float
            Energy of the system.
        """

        if check_input:
            directions = _check_directions(self._I, directions)
            cone_axis = _check_cone_axis(cone_axis=cone_axis)

        energy = 0

        T = np.eye(3, dtype=float) - np.outer(cone_axis, cone_axis)
        P = np.array(
            [
                [0, cone_axis[2], -cone_axis[1]],
                [-cone_axis[2], 0, cone_axis[0]],
                [cone_axis[1], -cone_axis[0], 0],
            ],
            dtype=float,
        )

        for i, j, R, J in self._bonds:
            sin_qr = np.sin(2 * np.pi * (self.spiral_vector @ R))
            cos_qr = np.cos(2 * np.pi * (self.spiral_vector @ R))
            T_nu = -sin_qr * P + cos_qr * T
            P_nu = sin_qr * T + cos_qr * P
            energy += (
                0.5
                * self._spins[i]
                * self._spins[j]
                * (
                    (cone_axis @ directions[i])
                    * (cone_axis @ J.matrix @ cone_axis)
                    * (cone_axis @ directions[j])
                    + (
                        np.einsum(
                            "i,j,kl,ki,lj",
                            directions[i],
                            directions[j],
                            J.matrix,
                            T,
                            T_nu,
                        )
                    )
                )
            )

        # Compute zeeman energy
        if self.magnetic_field is not None:
            hn = self.magnetic_field @ cone_axis
            for i in range(self._I):
                energy += (
                    BOHR_MAGNETON
                    * self._g_factors[i]
                    * hn
                    * self._spins[i]
                    * (cone_axis @ directions[i])
                )

        return energy

    def gradient(self, directions, cone_axis, check_input=True) -> np.ndarray:
        r"""
        Compute the gradient of the energy with respect to the provided directions of
        spins. Computed according to the formula FIXME of FIXME.

        Parameters
        ----------
        directions : (I, 3) or (3,) |array-like|_
            Orientation of the spin vectors.
            If ``I = 1``, then both ``(1,3)`` and ``(3,)`` shaped inputs are accepted.
            The vectors are normalized to one, i.e. only the direction of the vectors
            is important.
        cone_axis : (3,) |array-like|_
            Cone axis of the spiral cone state.
        check_input : bool, default=True
            Whether to check the correctness of the input. Better keep it ``True``
            always, unless you need to compute energy thousands of times and you control
            the input correctness externally.

        Returns
        -------
        gradient : (I,3) :numpy:`ndarray`
            Gradient of the energy. Form:

            .. code-block:: python

                [
                    [ dE/de1x, dE/de1y, dE/de1z ],
                    [ dE/de2x, dE/de2y, dE/de2z ],
                    ...
                    [ dE/deIx, dE/deIy, dE/deIz ],
                    [dE/dnx, dE/dny, dE/dnz]
                ]
        """

        if check_input:
            directions = _check_directions(self._I, directions)
            cone_axis = _check_cone_axis(cone_axis=cone_axis)

        gradient = np.zeros((self._I + 1, 3), dtype=float)

        T = np.eye(3, dtype=float) - np.outer(cone_axis, cone_axis)
        P = np.array(
            [
                [0, cone_axis[2], -cone_axis[1]],
                [-cone_axis[2], 0, cone_axis[0]],
                [cone_axis[1], -cone_axis[0], 0],
            ],
            dtype=float,
        )
        for i, j, R, J in self._bonds:
            sin_qr = np.sin(2 * np.pi * (self.spiral_vector @ R))
            cos_qr = np.cos(2 * np.pi * (self.spiral_vector @ R))
            T_nu = -sin_qr * P + cos_qr * T
            P_nu = sin_qr * T + cos_qr * P

            gradient[j] += (
                self._spins[i]
                * self._spins[j]
                * (
                    (cone_axis @ directions[i])
                    * (cone_axis @ J.matrix @ cone_axis)
                    * cone_axis
                    + cos_qr
                    * (np.einsum("i,kl,ki,lx->x", directions[i], J.matrix, T, T))
                )
            )

            gradient[-1] += (
                self._spins[i]
                * self._spins[j]
                * np.einsum(
                    "k,kx->x",
                    (
                        cone_axis * (cone_axis @ directions[i])
                        - cos_qr * np.einsum("ki,i->k", T, directions[i])
                    ),
                    (
                        np.einsum("x,kj,j->kx", directions[j], J.matrix, cone_axis)
                        + np.einsum("j,kx,j->kx", directions[j], J.matrix, cone_axis)
                    ),
                )
            )

        # Zeeman
        if self.magnetic_field is not None:
            hn = self.magnetic_field @ cone_axis
            for i in range(self._I):
                gradient[i] = (
                    BOHR_MAGNETON * self._g_factors[i] * self._spins * hn * cone_axis
                )

                gradient[-1] += (
                    BOHR_MAGNETON
                    * self._g_factors[i]
                    * self._spins[i]
                    * (
                        self.magnetic_field * (cone_axis @ directions[i])
                        + directions[i] * hn
                    )
                )

        return gradient

    def _F(self, true_variables) -> float:
        # Remove input check as this function is specific for the minimization procedure
        return self.energy(
            directions=true_variables[0],
            cone_axis=true_variables[1],
            check_input=False,
        )

    def _grad_F(self, true_variables) -> np.ndarray:
        r"""
        Compute the gradient of the energy as a function of
        :math:`(a^x_1, a^y_1, a^z_1, ..., a^x_I, a^y_I, a^z_I)` for the C1 sub-type.

        Parameters
        ----------
        true_variables : (I, 3) :numpy:`ndarray`
            Directions of the spin vectors.

        Returns
        -------
        gradient : (I*3,) :numpy:`ndarray`
            Gradient of the energy. Form:

            .. code-block:: python

                [ dF/da1x, dF/da1y, dF/da1z, ..., dF/daIx, dF/daIy, dF/daIz ]
        """

        directions = true_variables[0]
        cone_axis = true_variables[1]

        # Remove input check as this function is specific for the minimization procedure
        gradient = self.gradient(directions, cone_axis, check_input=False)

        torque_directions = np.cross(directions, gradient[:-1])
        torque_cone_axis = np.cross(cone_axis, gradient[-1])

        new_grad = np.zeros(3 * gradient.shape[0], dtype=float)
        new_grad[:-3] = torque_directions.flatten()
        new_grad[-3:] = torque_cone_axis

        return new_grad

    def _to_x(self, true_variables):
        r"""

        Parameters
        ----------
        true_variables : (I, 3) :numpy:`ndarray`
            Original direction of the spin vectors.

        Returns
        -------
        x : ((I+2)*3,) :numpy:`ndarray`
            Direction of the spin vectors parameterized with the skew-symmetric matrix.
        """

        return np.zeros(true_variables[0].size + 3, dtype=float)

    def _update(self, true_variables, x):
        r"""

        Parameters
        ----------
        true_variables : (I, 3) :numpy:`ndarray`
            Original direction of the spin vectors.
        x : (I*3,) :numpy:`ndarray`
            Direction of the spin vectors parameterized with the skew-symmetric matrix.

        Returns
        -------
        directions : (I, 3) :numpy:`ndarray`
            Rotated set of direction vectors.
        """

        directions = true_variables[0].copy()

        ax = x[:-6:3]
        ay = x[1:-6:3]
        az = x[2:-6:3]

        thetas = np.sqrt(ax**2 + ay**2 + az**2)

        for i in range(len(thetas)):
            theta = thetas[i]

            if theta < np.finfo(float).eps:
                continue

            r = np.array([ax[i], ay[i], az[i]]) / theta

            directions[i] = (
                np.cos(theta) * directions[i]
                + np.sin(theta) * np.cross(r, directions[i])
                + (1 - np.cos(theta)) * r * (r @ directions[i])
            )

        cone_axis = true_variables[1].copy()

        ax = x[-6]
        ay = x[-5]
        az = x[-4]

        theta = np.sqrt(ax**2 + ay**2 + az**2)

        if not (theta < np.finfo(float).eps):
            r = np.array([ax, ay, az]) / theta

            cone_axis = (
                np.cos(theta) * cone_axis
                + np.sin(theta) * np.cross(r, cone_axis)
                + (1 - np.cos(theta)) * r * (r @ cone_axis)
            )

        return [directions, cone_axis]

    def _update_difference(self, func_k, func_kp1, grad_kp1):
        r"""
        Computes the difference between two consecutive steps of the optimization.

        Parameters
        ----------
        func_k : float
            Energy at the current step.
        func_kp1 : float
            Energy at the next step.
        grad_kp1 : (I*3,) :numpy:`ndarray`
            Gradient of the energy at the next state.

        Returns
        -------
        difference : (2,) :numpy:`ndarray`
            Difference between the two consecutive steps of the optimization.
        """

        grad_directions = grad_kp1[:-6]
        grad_cone_axis = grad_kp1[-6:-3]
        return np.array(
            [
                abs(func_kp1 - func_k),
                np.linalg.norm(
                    grad_directions.reshape(grad_directions.size // 3, 3), axis=1
                ).max(),
                np.linalg.norm(grad_cone_axis),
            ]
        )

    def optimize(
        self,
        directions_ig=None,
        cone_axis_ig=None,
        energy_tolerance=1e-5,
        torque_tolerance=1e-5,
        ca_torque_tolerance=1e-5,
    ):
        r"""
        Optimize the energy with respect to the directions of spins in the unit cell.

        Parameters
        ----------
        directions_ig : (I, 3) or (3,) |array-like|_, optional
            Initial guess for the direction of the spin vectors.
        energy_tolerance : float, default 1e-5
            Energy tolerance for the two consecutive steps of the optimization.
        torque_tolerance : float, default 1e-5
            Torque tolerance for the two consecutive steps of the optimization.


        Returns
        -------
        optimized_directions : (I, 3) :numpy:`ndarray`
            Optimized direction of the spin vectors.
        """

        if directions_ig is None:
            directions_ig = np.random.uniform(low=-1, high=1, size=(self._I, 3))
            _logger.debug("Initial guess for the optimization is random (C5).")

        if cone_axis_ig is None:
            cone_axis_ig = np.random.uniform(low=-1, high=1, size=3)

        # Check the input correctness and normalize vectors.
        directions = _check_directions(self._I, directions_ig)

        cone_axis = _check_cone_axis(cone_axis=cone_axis_ig)

        optimized_directions = _optimize(
            self,
            true_variables_0=[directions, cone_axis],
            tolerance=np.array(
                [energy_tolerance, torque_tolerance, ca_torque_tolerance]
            ),
            func=self._F,
            grad=self._grad_F,
            update=self._update,
            to_x=self._to_x,
            update_difference=self._update_difference,
        )

        return optimized_directions


class C5(_Energy):
    def __init__(self, spinham: SpinHamiltonian):
        super().__init__(spinham)

    def energy(self, directions, cone_axis, spiral_vector, check_input=True) -> float:
        r"""
        Compute the energy of the system with respect to the provided directions of
        spins, cone axis and spiral vector. Computed according to the formula FIXME of FIXME.

        Parameters
        ----------
        directions : (I, 3) or (3,) |array-like|_
            Orientation of the spin vectors.
            If ``I = 1``, then both ``(1,3)`` and ``(3,)`` shaped inputs are accepted.
            The vectors are normalized to one, i.e. only the direction of the vectors
            is important.
        cone_axis : (3,) |array-like|_
            Cone axis of the spiral cone state.
        spiral_vector : (3,) |array-like|_
            Spiral vector of the spiral cone state.
            Relative to the three reciprocal lattice vectors.
        check_input : bool, default=True
            Whether to check the correctness of the input. Better keep it ``True``
            always, unless you need to compute energy thousands of times and you control
            the input correctness externally.

        Returns
        -------
        energy : float
            Energy of the system.
        """

        if check_input:
            directions = _check_directions(self._I, directions)
            cone_axis = _check_cone_axis(cone_axis=cone_axis)
            spiral_vector = _check_spiral_vector(spiral_vector=spiral_vector)

        energy = 0

        T = np.eye(3, dtype=float) - np.outer(cone_axis, cone_axis)
        P = np.array(
            [
                [0, cone_axis[2], -cone_axis[1]],
                [-cone_axis[2], 0, cone_axis[0]],
                [cone_axis[1], -cone_axis[0], 0],
            ],
            dtype=float,
        )

        for i, j, R, J in self._bonds:
            sin_qr = np.sin(2 * np.pi * (spiral_vector @ R))
            cos_qr = np.cos(2 * np.pi * (spiral_vector @ R))
            T_nu = -sin_qr * P + cos_qr * T
            P_nu = sin_qr * T + cos_qr * P
            energy += (
                0.5
                * self._spins[i]
                * self._spins[j]
                * (
                    (cone_axis @ directions[i])
                    * (cone_axis @ J.matrix @ cone_axis)
                    * (cone_axis @ directions[j])
                    + 0.5
                    * (
                        np.einsum(
                            "i,j,kl,ki,lj",
                            directions[i],
                            directions[j],
                            J.matrix,
                            T,
                            T_nu,
                        )
                        + np.einsum(
                            "i,j,kl,ki,lj",
                            directions[i],
                            directions[j],
                            J.matrix,
                            P,
                            P_nu,
                        )
                    )
                )
            )

        # Compute zeeman energy
        if self.magnetic_field is not None:
            hn = self.magnetic_field @ cone_axis
            for i in range(self._I):
                energy += (
                    BOHR_MAGNETON
                    * self._g_factors[i]
                    * hn
                    * self._spins[i]
                    * (cone_axis @ directions[i])
                )

        return energy

    def gradient(
        self, directions, cone_axis, spiral_vector, check_input=True
    ) -> np.ndarray:
        r"""
        Compute the gradient of the energy with respect to the provided directions of
        spins. Computed according to the formula FIXME of FIXME.

        Parameters
        ----------
        directions : (I, 3) or (3,) |array-like|_
            Orientation of the spin vectors.
            If ``I = 1``, then both ``(1,3)`` and ``(3,)`` shaped inputs are accepted.
            The vectors are normalized to one, i.e. only the direction of the vectors
            is important.
        cone_axis : (3,) |array-like|_
            Cone axis of the spiral cone state.
        spiral_vector : (3,) |array-like|_
            Spiral vector of the spiral cone state.
            Relative to the three reciprocal lattice vectors.
        check_input : bool, default=True
            Whether to check the correctness of the input. Better keep it ``True``
            always, unless you need to compute energy thousands of times and you control
            the input correctness externally.

        Returns
        -------
        gradient : (I,3) :numpy:`ndarray`
            Gradient of the energy. Form:

            .. code-block:: python

                [
                    [ dE/de1x, dE/de1y, dE/de1z ],
                    [ dE/de2x, dE/de2y, dE/de2z ],
                    ...
                    [ dE/deIx, dE/deIy, dE/deIz ],
                    [dE/dnx, dE/dny, dE/dnz],
                    [dE/dqx, dE/dqy, dE/dqz],
                ]
        """

        if check_input:
            directions = _check_directions(self._I, directions)
            cone_axis = _check_cone_axis(cone_axis=cone_axis)
            spiral_vector = _check_spiral_vector(spiral_vector=spiral_vector)

        gradient = np.zeros((self._I + 2, 3), dtype=float)

        T = np.eye(3, dtype=float) - np.outer(cone_axis, cone_axis)
        P = np.array(
            [
                [0, cone_axis[2], -cone_axis[1]],
                [-cone_axis[2], 0, cone_axis[0]],
                [cone_axis[1], -cone_axis[0], 0],
            ],
            dtype=float,
        )
        for i, j, R, J in self._bonds:
            sin_qr = np.sin(2 * np.pi * (spiral_vector @ R))
            cos_qr = np.cos(2 * np.pi * (spiral_vector @ R))
            T_nu = -sin_qr * P + cos_qr * T
            P_nu = sin_qr * T + cos_qr * P

            gradient[j] += (
                self._spins[i]
                * self._spins[j]
                * (
                    (cone_axis @ directions[i])
                    * (cone_axis @ J.matrix @ cone_axis)
                    * cone_axis
                    + 0.5
                    * (
                        np.einsum("i,kl,ki,lx->x", directions[i], J.matrix, T, T_nu)
                        + np.einsum("i,kl,ki,lx->x", directions[i], J.matrix, P, P_nu)
                    )
                )
            )

            gradient[-2] += (
                self._spins[i]
                * self._spins[j]
                * (
                    np.einsum(
                        "xl,l->x",
                        (
                            np.einsum("x,k,kl->xl", directions[i], cone_axis, J.matrix)
                            + np.einsum(
                                "k,k,xl->xl", directions[i], cone_axis, J.matrix
                            )
                        ),
                        (
                            cone_axis * (cone_axis @ directions[j])
                            - 0.5 * np.einsum("lj,j->l", T_nu, directions[j])
                        ),
                    )
                    - np.cross(
                        directions[i],
                        np.einsum("kl,lj,j -> k", J.matrix, P_nu, directions[j]),
                    )
                )
            )

            gradient[-1] += (
                0.25
                * self._spins[i]
                * self._spins[j]
                * (R @ self._cell)
                * (
                    np.einsum(
                        "i,j,kl,ki,lj", directions[i], directions[j], J.matrix, P, T_nu
                    )
                    - np.einsum(
                        "i,j,kl,ki,lj", directions[i], directions[j], J.matrix, T, P_nu
                    )
                )
            )

        # Zeeman
        if self.magnetic_field is not None:
            hn = self.magnetic_field @ cone_axis
            for i in range(self._I):
                gradient[i] = (
                    BOHR_MAGNETON * self._g_factors[i] * self._spins * hn * cone_axis
                )

                gradient[-2] += (
                    BOHR_MAGNETON
                    * self._g_factors[i]
                    * self._spins[i]
                    * (
                        self.magnetic_field * (cone_axis @ directions[i])
                        + directions[i] * hn
                    )
                )

        return gradient

    def _F(self, true_variables) -> float:
        # Remove input check as this function is specific for the minimization procedure
        return self.energy(
            directions=true_variables[0],
            cone_axis=true_variables[1],
            spiral_vector=true_variables[2],
            check_input=False,
        )

    def _grad_F(self, true_variables) -> np.ndarray:
        r"""
        Compute the gradient of the energy as a function of
        :math:`(a^x_1, a^y_1, a^z_1, ..., a^x_I, a^y_I, a^z_I)` for the C1 sub-type.

        Parameters
        ----------
        true_variables : (I, 3) :numpy:`ndarray`
            Directions of the spin vectors.

        Returns
        -------
        gradient : (I*3,) :numpy:`ndarray`
            Gradient of the energy. Form:

            .. code-block:: python

                [ dF/da1x, dF/da1y, dF/da1z, ..., dF/daIx, dF/daIy, dF/daIz ]
        """

        directions = true_variables[0]
        cone_axis = true_variables[1]
        spiral_vector = true_variables[2]

        # Remove input check as this function is specific for the minimization procedure
        gradient = self.gradient(
            directions, cone_axis, spiral_vector, check_input=False
        )

        torque_directions = np.cross(directions, gradient[:-2])
        torque_cone_axis = np.cross(cone_axis, gradient[-2])

        new_grad = np.zeros(3 * gradient.shape[0], dtype=float)
        new_grad[:-6] = torque_directions.flatten()
        new_grad[-6:-3] = torque_cone_axis
        new_grad[-3:] = gradient[-1]

        return new_grad

    def _to_x(self, true_variables):
        r"""

        Parameters
        ----------
        true_variables : (I, 3) :numpy:`ndarray`
            Original direction of the spin vectors.

        Returns
        -------
        x : ((I+2)*3,) :numpy:`ndarray`
            Direction of the spin vectors parameterized with the skew-symmetric matrix.
        """

        return np.concatenate(
            (np.zeros(true_variables[0].size + 3, dtype=float), true_variables[-1])
        )

    def _update(self, true_variables, x):
        r"""

        Parameters
        ----------
        true_variables : (I, 3) :numpy:`ndarray`
            Original direction of the spin vectors.
        x : (I*3,) :numpy:`ndarray`
            Direction of the spin vectors parameterized with the skew-symmetric matrix.

        Returns
        -------
        directions : (I, 3) :numpy:`ndarray`
            Rotated set of direction vectors.
        """

        directions = true_variables[0].copy()

        ax = x[:-6:3]
        ay = x[1:-6:3]
        az = x[2:-6:3]

        thetas = np.sqrt(ax**2 + ay**2 + az**2)

        for i in range(len(thetas)):
            theta = thetas[i]

            if theta < np.finfo(float).eps:
                continue

            r = np.array([ax[i], ay[i], az[i]]) / theta

            directions[i] = (
                np.cos(theta) * directions[i]
                + np.sin(theta) * np.cross(r, directions[i])
                + (1 - np.cos(theta)) * r * (r @ directions[i])
            )

        cone_axis = true_variables[1].copy()

        ax = x[-6]
        ay = x[-5]
        az = x[-4]

        theta = np.sqrt(ax**2 + ay**2 + az**2)

        if not (theta < np.finfo(float).eps):
            r = np.array([ax, ay, az]) / theta

            cone_axis = (
                np.cos(theta) * cone_axis
                + np.sin(theta) * np.cross(r, cone_axis)
                + (1 - np.cos(theta)) * r * (r @ cone_axis)
            )

        spiral_vector = true_variables[2] + x[-3:]

        return [directions, cone_axis, spiral_vector]

    def _update_difference(self, func_k, func_kp1, grad_kp1):
        r"""
        Computes the difference between two consecutive steps of the optimization.

        Parameters
        ----------
        func_k : float
            Energy at the current step.
        func_kp1 : float
            Energy at the next step.
        grad_kp1 : (I*3,) :numpy:`ndarray`
            Gradient of the energy at the next state.

        Returns
        -------
        difference : (2,) :numpy:`ndarray`
            Difference between the two consecutive steps of the optimization.
        """

        grad_directions = grad_kp1[:-6]
        grad_cone_axis = grad_kp1[-6:-3]
        return np.array(
            [
                abs(func_kp1 - func_k),
                np.linalg.norm(
                    grad_directions.reshape(grad_directions.size // 3, 3), axis=1
                ).max(),
                np.linalg.norm(grad_cone_axis),
            ]
        )

    def optimize(
        self,
        directions_ig=None,
        cone_axis_ig=None,
        spiral_vector_ig=None,
        energy_tolerance=1e-5,
        torque_tolerance=1,
        ca_torque_tolerance=1,
    ):
        r"""
        Optimize the energy with respect to the directions of spins in the unit cell.

        Parameters
        ----------
        directions_ig : (I, 3) or (3,) |array-like|_, optional
            Initial guess for the direction of the spin vectors.
        energy_tolerance : float, default 1e-5
            Energy tolerance for the two consecutive steps of the optimization.
        torque_tolerance : float, default 1e-5
            Torque tolerance for the two consecutive steps of the optimization.


        Returns
        -------
        optimized_directions : (I, 3) :numpy:`ndarray`
            Optimized direction of the spin vectors.
        """

        if directions_ig is None:
            directions_ig = np.random.uniform(low=-1, high=1, size=(self._I, 3))
            _logger.debug("Initial guess for the optimization is random (C5).")

        if cone_axis_ig is None:
            cone_axis_ig = np.random.uniform(low=-1, high=1, size=3)

        if spiral_vector_ig is None:
            spiral_vector_ig = np.random.uniform(low=-1, high=1, size=3)

        # Check the input correctness and normalize vectors.
        directions = _check_directions(self._I, directions_ig)

        cone_axis = _check_cone_axis(cone_axis=cone_axis_ig)
        spiral_vector = _check_spiral_vector(spiral_vector=spiral_vector_ig)

        optimized_directions = _optimize(
            self,
            true_variables_0=[directions, cone_axis, spiral_vector],
            tolerance=np.array(
                [energy_tolerance, torque_tolerance, ca_torque_tolerance]
            ),
            func=self._F,
            grad=self._grad_F,
            update=self._update,
            to_x=self._to_x,
            update_difference=self._update_difference,
        )

        return optimized_directions



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging

_all_ = []

_logger = logging.getLogger(__name__)


class _FailedToVerifyTxtModelFile(Exception):
    R"""
    Raise if the format of the model input file is invalid.
    """

    def __init__(self, message):
        super().__init__(message)


TRUE_KEYWORDS = ["true", "t", "yes", "y", "1"]
FALSE_KEYWORDS = ["false", "f", "no", "n", "0"]


################################################################################
#                              Data type checkers                              #
################################################################################
def _is_float(word) -> bool:
    r"""
    Check if the ``word`` can be converted to ``float``

    Returns
    -------
    bool
        Whether the word can be converted to a float.
    """
    try:
        word = float(word)
        return True
    except ValueError:
        return False


def _is_integer(word) -> bool:
    r"""
    Check if the ``word`` can be converted to ``int``

    Returns
    -------
    bool
        Whether the word can be converted to an integer.
    """

    try:
        word = int(word)
        return True
    except ValueError:
        return False


def _is_bool(word) -> bool:
    r"""
    Check if the ``word`` is one of the supported keywords for boolean values.

    Returns
    -------
    bool
        Whether the word is one of the supported keywords for boolean values.
    """
    return word.lower() in TRUE_KEYWORDS + FALSE_KEYWORDS


################################################################################
#                                Common checker                                #
################################################################################
def _verify_atom_name(word, line_index) -> bool:
    r"""
    Check if the ``word`` is a valid atom's name.

    Parameters
    ----------
    word : str
        The word to check.
    line_index : int
        Original line number, before filtering.

    Returns
    -------
    errors : bool
        ``True`` if errors are found, ``False`` otherwise.
    """
    # Atom's name has to start from a number or a letter and should not contain any "#".abs
    # Atom's name should not start nor end with double underscore "__"
    if (
        not (str.isalnum(word[0]) and word.count("#") == 0)
        or word.startswith("__")
        or word.endswith("__")
    ):
        _logger.error(
            " ".join(
                [
                    f"Line {line_index}: Atom names have to start with a",
                    'letter or a number and should not contain any "#" symbols',
                    'and should not start nor end with double underscore "__",',
                    f'got "{word}"',
                ]
            )
        )
        return True
    return False


def _verify_section_header(line, line_index, units_names) -> bool:
    r"""
    Check the section header.

    Parameters
    ----------
    line : str
        The first line of the section.
    line_index : int
        Original line number, before filtering.
    units_names : list of str
        List of the supported units names for the section.
        It is check if the units start with one of the names from the list.

    Returns
    -------
    errors : bool
        ``True`` if errors are found, ``False`` otherwise.

    Notes
    -----
    The section name is checked at the other place.
    """
    # There are no errors yet
    errors = False

    # Checker for the atom's coordinate units
    def is_units_keyword(word):
        for entry in units_names:
            if word.lower().startswith(entry):
                return True

        return False

    # Both section name and units are case insensitive
    line = line.lower().split()
    section_name = line[0]

    # If <units> are present, then line has 2 blocks:
    # Section_name <units>
    if len(line) == 2:
        # Check that units are expected
        if len(units_names) == 0:
            errors = True
            _logger.error(
                " ".join(
                    [
                        f"Line {line_index}: expected only the section name",
                        f'and no <units>, got "{" ".join(line)}".',
                    ]
                )
            )
        # If units are expected check that they are one of the supported ones
        elif not is_units_keyword(line[1]):
            errors = True
            _logger.error(
                " ".join(
                    [
                        f"Line {line_index}, block 2: expected word starting from",
                        ", ".join([f'"{i}"' for i in units_names]),
                        f'got "{line[1]}".',
                    ]
                )
            )
    # If <units> are not present, then the line has only one block:
    # section_name
    elif len(line) != 1:
        # If optional units are expected
        if len(units_names) > 0:
            errors = True
            _logger.error(
                " ".join(
                    [
                        f'Line {line_index}: expected "{section_name}" keyword',
                        f'and/or <units>, got "{" ".join(line)}".',
                    ]
                )
            )
        # If no units are expected
        else:
            errors = True
            _logger.error(
                " ".join(
                    [
                        f'Line {line_index}: expected "{section_name}" keyword',
                        f'and no <units>, got "{" ".join(line)}".',
                    ]
                )
            )

    return errors


def _verify_numerical_data_block(lines, data_specs, line_indices, keyword=None) -> bool:
    r"""
    Verify one data block.

    Data block starts with the keyword (only the first letter is compared) and might be followed
    by any number of lines with numbers. ``data_specs`` is a list of amount of numbers on each line.
    ``data_specs[0]`` - amount of numbers in the same line as the keyword.
    ``data_specs[1:]`` - amount of numbers in the following lines.

    Parameters
    ----------
    lines : list of str
        List of the data block lines from the input file.
        ``len(lines) == len(line_indices) == len(data_specs) > 0``.
    data_specs : list of (int or list of int)
        List of the amount of numbers on each line.
        ``len(data_specs) == len(line_indices) == len(lines) > 0``.
        if ``data_specs[i]`` is a list, then the line has to have one of the amounts of numbers
        from the list.
    line_indices : list of int
        Original line numbers, before filtering.
        ``len(line_indices) == len(data_specs) == len(lines) > 0``.
    keyword : str, optional
        The keyword of the data block. Case insensitive.

    Returns
    -------
    errors : bool
        ``True`` if errors are found, ``False`` otherwise.
    """

    # There are no errors yet
    errors = False

    # Check the amount of lines
    if len(lines) != len(data_specs):
        errors = True
        _logger.error(
            " ".join(
                [
                    f"Line {line_indices[0]}: Expected to have {len(data_specs)} lines,"
                    f"got {len(lines)} lines",
                ]
            )
        )

    # Check that every line
    for l_i, line in enumerate(lines):
        line = line.lower().split()

        keyword_error_addition = ""

        # If the keyword is present in the first line
        if l_i == 0 and keyword is not None:
            # Check that the keyword is the expected one
            if keyword.lower()[0] != line[0][0]:
                errors = True
                _logger.error(
                    " ".join(
                        [
                            f"Line {line_indices[l_i]}: expected to have the",
                            f'"{keyword}" keyword (only first letter is checked),'
                            f'got "{line[0][0]}"',
                        ]
                    )
                )
            # Remove the keyword from the line
            line = line[1:]
            # Add the keyword to the error message
            keyword_error_addition = f'"{keyword}" keyword followed by'

        # Check the amount of numbers in the line
        if isinstance(data_specs[l_i], int):
            amount_is_correct = len(line) == data_specs[l_i]
            possible_amounts = str(data_specs[l_i])
        else:
            amount_is_correct = False
            for amount in data_specs[l_i]:
                if len(line) == amount:
                    amount_is_correct = True
                    break
            possible_amounts = " or ".join([str(i) for i in data_specs[l_i]])

        if not amount_is_correct:
            errors = True
            _logger.error(
                " ".join(
                    [
                        f"Line {line_indices[l_i]}: expected to have",
                        keyword_error_addition,
                        f"{possible_amounts} numbers, got {len(line)} numbers",
                    ]
                )
            )
        # Check that every number is convertible to a float
        for n_i, number in enumerate(line):
            if not _is_float(number):
                errors = True
                _logger.error(
                    " ".join(
                        [
                            f"Line {line_indices[l_i]}, block {n_i+1}: expected a number,",
                            f'got "{number}"',
                        ]
                    )
                )

    return errors


def _verify_size_of_section(
    section_name, lines, line_index, expected_lines, exact=True
) -> bool:
    r"""
    Verify the size of the section.

    Parameters
    ----------
    section_name : str
        Name of the section. Is used just for the error message, nothing is checked.
    lines : list of str
        List of the section lines from the input file.
        Without comments and blank lines.
    line_index : int
        Original line number of the first line of the section, before filtering.
    expected_lines : int or list of int
        Expected amount of lines in the section. If list, then the section has to have
        one of the amounts of lines from the list.
    exact : bool, default True
        Whether the section has to have exactly ``expected_lines`` lines.

        * If True, then the section has to have exactly ``expected_lines`` lines.
        * If False, then the section has to have at least ``expected_lines`` lines.

    Returns
    -------
    errors : bool
        ``True`` if errors are found, ``False`` otherwise.
    """

    # There are no errors yet
    errors = False

    # Check that the section has an exact amount of lines
    if exact:
        # If only one option is provided
        if isinstance(expected_lines, int):
            amount_is_correct = len(lines) == expected_lines
            possible_amounts = str(expected_lines)
        # If multiple options are provided
        else:
            amount_is_correct = False
            for amount in expected_lines:
                if len(lines) == amount:
                    amount_is_correct = True
                    break
            possible_amounts = " or ".join([str(i) for i in expected_lines])
        # If the amount is not correct
        if not amount_is_correct:
            errors = True
            _logger.error(
                " ".join(
                    [
                        f"Line {line_index}: expected to have {section_name} section with",
                        f"exactly {possible_amounts} lines, got {len(lines)} lines",
                    ]
                )
            )
    # Check that the section has at least the expected amount of lines
    else:
        # Meaningful error message for the programmer
        if not isinstance(expected_lines, int):
            raise RuntimeError(
                "_verify_size_of_section: If exact is False, then expected_lines has to be an integer"
            )
        if len(lines) < expected_lines:
            errors = True
            _logger.error(
                " ".join(
                    [
                        f"Line {line_index}: expected to have {section_name} section with",
                        f"at least {expected_lines} lines, got {len(lines)} lines",
                    ]
                )
            )

    return errors


def _verify_short_separator(line, line_index, symbol) -> bool:
    r"""
    Check if the separator is too short.

    Parameters
    ----------
    line : str
        The line to check.
    line_index : int
        Original line number, before filtering.
    symbol : str
        The symbol of the separator.

    Returns
    -------
    errors : bool
        ``True`` if errors are found, ``False`` otherwise.
    """
    # There are no errors yet
    errors = False
    if len(line.strip()) < 10 and len(line.strip()) == line.count(symbol):
        errors = True
        _logger.error(
            " ".join(
                [
                    f"Line {line_index}: Potential separator is too short",
                    f'(expected 10 or more "{symbol}" symbols),',
                    f'got {line.count(symbol)} "{symbol}" symbols.',
                ]
            )
        )
    return errors


################################################################################
#                             Cell section checker                             #
################################################################################
def _verify_cell(lines, line_indices) -> bool:
    r"""
    Check that the found "cell" section is following the input file specification.

    Parameters
    ----------
    lines : list of str
        List of the "cell" section lines from the input file.
        Without comments and blank lines.
        ``len(lines) == len(line_indices)``
    line_indices : list of int
        Original line numbers, before filtering.
        ``len(line_indices) == len(lines)``

    Returns
    -------
    errors : bool
        ``True`` if errors are found, ``False`` otherwise.
    """
    # At the beginning we assume that the first line starts with the
    # case-insensitive word "cell", followed by the next line symbol or a space.

    # There are no errors yet
    errors = False

    # Check size of the cell section it has to have exactly 4 or 5 lines
    if _verify_size_of_section("cell", lines, line_indices[0], [4, 5], exact=True):
        # Do not proceed with the rest of the checks,
        # since the behavior of the rest of the checks is unpredictable
        return True

    # Starting from this line it is assumed that the section has exactly 4 or 5 lines

    # Check the section header
    errors = errors or _verify_section_header(lines[0], line_indices[0], ["a", "b"])

    # If scale factor is present
    if len(lines) == 5:
        scale_errors = _verify_numerical_data_block(
            [lines[1]], [[1, 3]], [line_indices[1]]
        )
        errors = errors or scale_errors
        scale = lines[1].split()
        if not scale_errors and len(scale) == 1 and float(scale[0]) == 0:
            errors = True
            _logger.error("Scale can not be zero")
        if (
            not scale_errors
            and len(scale) == 3
            and (float(scale[0]) <= 0 or float(scale[1]) <= 0 or float(scale[2]) <= 0)
        ):
            errors = True
            _logger.error(
                f'Scale as three numbers - all have to be positive, got "{scale}"'
            )

        cell_lines = lines[2:]
        cell_lines_indices = line_indices[2:]
    # If no scale factor is present
    else:
        cell_lines = lines[1:]
        cell_lines_indices = line_indices[1:]

    # Check that every lattice vector is provided as three numbers separated by spaces.
    errors = errors or _verify_numerical_data_block(
        cell_lines, [3, 3, 3], cell_lines_indices
    )

    return errors


################################################################################
#                            Atoms section checker                             #
################################################################################
def _verify_vector_keyword(keywords, liter, line_index) -> bool:
    r"""
    Check that one of the three sets is provided:

    * s
    * sx sy sz
    * st sp s

    Parameters
    ----------
    keywords : list of str
        List of the keywords.
    liter : int
        liter of the value (i.e. "s", "l" or "j").
    line_index : int
        Original line number, before filtering.

    Returns
    -------
    errors : bool
        ``True`` if errors are found, ``False`` otherwise.
    """

    # Sorted keywords
    # <liter>
    # or
    # <liter>x <liter>y <liter>z
    # or
    # <liter> <liter>p <liter>t
    # or nothing
    # or any combination of those variants that keeps unique keywords
    keywords.sort()

    if not (
        keywords == [f"{liter}x", f"{liter}y", f"{liter}z"]
        or keywords == [liter, f"{liter}p", f"{liter}t"]
        or keywords == [liter]
        or keywords == [liter, f"{liter}x", f"{liter}y", f"{liter}z"]
        or keywords
        == [liter, f"{liter}p", f"{liter}t", f"{liter}x", f"{liter}y", f"{liter}z"]
        or not keywords
    ):
        _logger.error(
            " ".join(
                [
                    f"Line {line_index}: expected to have full combination or",
                    "several full combinations of vector keywords or nothing",
                    f'got "{" ".join(keywords)}".',
                    "Check atoms section of documentation for the correct format.",
                ]
            )
        )
        return True

    return False


def _verify_atoms_data_header(line, line_index):
    r"""
    Check the data header of the "atoms" section.

    Parameters
    ----------
    line : str
        The second line of the "atoms" section.
    line_index : int
        Original line number, before filtering.

    Returns
    -------
    N : int
        Number of blocks in the data header.
    name_index : int or None
        Index of the block with the atom's name.
        If None, then the atom's name is not present.
    errors : bool
        ``True`` if errors are found, ``False`` otherwise.
    """

    # There are no errors yet
    errors = False
    name_index = None
    N = len(line.split())

    keywords = line.lower().split()

    # Check that every block of the header is unique:
    if not len(keywords) == len(set(keywords)):
        errors = True
        _logger.error(
            " ".join(
                [
                    f"Line {line_index}: expected unique blocks in the data header,",
                    f'got "{" ".join(keywords)}"',
                ]
            )
        )

    # Check that the required "name" keyword is present in the data header
    if "name" in keywords:
        name_index = keywords.index("name")
        keywords.remove("name")
    else:
        errors = True
        _logger.error(
            " ".join(
                [
                    f"Line {line_index}: expected to have the atom's name",
                    "in the data header, got none",
                ]
            )
        )

    # Remove charge keywords from the list
    if "q" in keywords:
        keywords.remove("q")
    # Remove g-factor keywords from the list
    if "g" in keywords:
        keywords.remove("g")

    position_keywords = []
    spin_keywords = []
    orbital_moment_keywords = []
    total_moment_keywords = []

    # Categorized keywords
    for keyword in keywords:
        if keyword.startswith("r") or keyword in "xyz":
            position_keywords.append(keyword)
        elif keyword.startswith("s"):
            spin_keywords.append(keyword)
        elif keyword.startswith("l"):
            orbital_moment_keywords.append(keyword)
        elif keyword.startswith("j"):
            total_moment_keywords.append(keyword)

    # Sort the lists
    position_keywords.sort()

    # Check position keywords
    if not (
        position_keywords == ["r1", "r2", "r3"]
        or position_keywords == ["x", "y", "z"]
        or position_keywords == ["r1", "r2", "r3", "x", "y", "z"]
    ):
        errors = True
        _logger.error(
            " ".join(
                [
                    f"Line {line_index}: expected to have three or six position keywords.",
                    f'Either "r1 r2 r3" or "x y z" or both, got "{" ".join(position_keywords)}"',
                ]
            )
        )
    # Remove position keywords from the list
    for keyword in position_keywords:
        keywords.remove(keyword)

    # Check spin keywords
    errors = errors or _verify_vector_keyword(spin_keywords, "s", line_index)
    # Remove spin keywords from the list
    for keyword in spin_keywords:
        keywords.remove(keyword)

    # Check orbital moment keywords
    errors = errors or _verify_vector_keyword(orbital_moment_keywords, "l", line_index)

    # Remove orbital moment keywords from the list
    for keyword in orbital_moment_keywords:
        keywords.remove(keyword)

    # Check total moment keywords
    errors = errors or _verify_vector_keyword(total_moment_keywords, "j", line_index)

    # Remove total moment keywords from the list
    for keyword in total_moment_keywords:
        keywords.remove(keyword)

    # Check if there are any unsupported keywords left
    if keywords:
        _logger.warning(
            " ".join(
                [
                    f"Line {line_index}: unsupported keywords in the data header:",
                    f'"{" ".join(keywords)}"',
                ]
            )
        )

    return N, name_index, errors


def _verify_atoms(lines, line_indices):
    r"""
    Check that the found "atoms" section is following the input file specification.

    Parameters
    ----------
    lines : list of str
        List of the "atoms" section lines from the input file.
        Without comments and blank lines.
        ``len(lines) == len(line_indices)``
    line_indices : list of int
        Original line numbers, before filtering.
        ``len(line_indices) == len(lines)``

    Returns
    -------
    allowed_atoms : set of str
        Set of the allowed atom's names.
    errors : bool
        ``True`` if errors are found, ``False`` otherwise.
    """

    allowed_atoms = set()
    # At the beginning we assume that the first line starts with the
    # case-insensitive word "atoms", followed by the next line symbol or a space.

    # There are no errors yet
    errors = False
    # Check size of the atoms section it has to have at least 3 lines: section header,
    # data header and at least one atom
    size_error = _verify_size_of_section(
        "atoms", lines, line_indices[0], 3, exact=False
    )
    if _verify_size_of_section("atoms", lines, line_indices[0], 3, exact=False):
        return set(), True

    # Starting from this line it is assumed that the section has at least 3 lines

    # Check the section header
    errors = errors or _verify_section_header(lines[0], line_indices[0], ["a", "b"])

    # Check the data header
    N, name_index, data_header_errors = _verify_atoms_data_header(
        lines[1], line_indices[1]
    )
    errors = errors or data_header_errors

    # Check each atom line
    for i in range(2, len(lines)):
        line = lines[i].split()

        # Check the amount of blocks
        if len(lines[i].split()) != N:
            errors = True
            _logger.error(
                " ".join(
                    [
                        f"Line {line_indices[i]}: expected to have {N} blocks",
                        f'as per data header, got "{lines[i]}"',
                    ]
                )
            )

        # Check atom's names if any
        if name_index is not None:
            errors = errors or _verify_atom_name(line[name_index], line_indices[i])
            allowed_atoms.add(line[name_index])

        # Check other data fields
        for b_i, block in enumerate(line):
            # Skip the name
            if name_index is not None and b_i == name_index:
                continue

            # All other blocks have to be numbers or "-" symbols (for missing data)
            if not _is_float(block) and block != "-":
                errors = True
                _logger.error(
                    " ".join(
                        [
                            f"Line {line_indices[i]}, block {b_i+1}: expected a number",
                            f'or "-" symbol, got "{block}"',
                        ]
                    )
                )

    if len(allowed_atoms) != len(lines) - 2:
        errors = True
        _logger.error(
            " ".join(
                [
                    f"Line {line_indices[2]}-{line_indices[-1]}: expected to have unique",
                    f"atom's names, got duplicates. {len(allowed_atoms)} unique names found.",
                ]
            )
        )

    return allowed_atoms, errors


################################################################################
#                          Convention section checker                          #
################################################################################
def _verify_notation(
    lines,
    line_indices,
    expect_exchange_factor=True,
    expect_on_site_factor=True,
    expect_double_counting=True,
    expect_spin_normalized=True,
) -> bool:
    r"""
    Check that the found "notation" section is following the input file specification.

    Parameters
    ----------
    lines : list of str
        List of the "notation" section lines from the input file.
        Without comments and blank lines.
        ``len(lines) == len(line_indices)``
    line_indices : list of int
        Original line numbers, before filtering.
        ``len(line_indices) == len(lines)``
    expect_exchange_factor : bool, default: True
        Whether the exchange factor is expected to be present.
    expect_on_site_factor : bool, default: True
        Whether the on-site factor is expected to be present.
    expect_double_counting : bool, default: True
        Whether the double-counting property is expected to be present.
    expect_spin_normalized : bool, default: True
        Whether the spin-normalized property is expected to be present.

    Returns
    -------
    errors : bool
        ``True`` if errors are found, ``False`` otherwise.
    """
    # There are no errors yet
    errors = False
    # Check condition about the size of the section the minimum amount of lines is
    # defined by expected keys
    errors = errors or _verify_size_of_section(
        "notation",
        lines,
        line_indices[0],
        int(expect_exchange_factor)
        + int(expect_on_site_factor)
        + int(expect_double_counting)
        + int(expect_spin_normalized)
        + 1,
        exact=False,
    )

    # Check the section header
    errors = errors or _verify_section_header(lines[0], line_indices[0], [])

    # Dictionary of the found properties
    found_properties = {}
    for i in range(1, len(lines)):
        line = lines[i].lower().split()
        if len(line) != 2:
            errors = True
            _logger.error(
                " ".join(
                    [
                        f"Line {line_indices[i]}: expected to have two blocks,",
                        f'separated by spaces, got "{lines[i]}"',
                    ]
                )
            )
        else:
            # Only first letter is checked
            if line[0][0] in found_properties:
                errors = True
                _logger.error(
                    " ".join(
                        [
                            f"Line {line_indices[i]}: found more than one entry of",
                            f'the "{line[0]}" property (only first letter is checked)',
                        ]
                    )
                )
            else:
                found_properties[line[0][0]] = (line[1], i)

    recognized_properties = ["d", "s", "e", "o"]
    expected = {
        "d": expect_double_counting,
        "s": expect_spin_normalized,
        "e": expect_exchange_factor,
        "o": expect_on_site_factor,
    }
    full_names = {
        "d": "double-counting",
        "s": "spin-normalized",
        "e": "exchange-factor",
        "o": "on-site-factor",
    }
    verify_functions = {
        "d": _is_bool,
        "s": _is_bool,
        "e": _is_float,
        "o": _is_float,
    }
    error_expect = {
        "d": "boolean",
        "s": "boolean",
        "e": "number",
        "o": "number",
    }

    # Check that every recognized property is present if it is expected
    # and that it has the correct value
    for prop in recognized_properties:
        # If the property is not found
        if prop not in found_properties:
            # But it is expected
            if expected[prop]:
                errors = True
                _logger.error(
                    f"Line {line_indices[0]}: did not "
                    f'find the "{full_names[prop]}" property in the notation section'
                )
        # If the property is found
        else:
            # Verify it
            if not verify_functions[prop](found_properties[prop][0]):
                errors = True
                _logger.error(
                    " ".join(
                        [
                            f"Line {line_indices[found_properties[prop][1]]}:",
                            f"expected to have a {error_expect[prop]},",
                            f'got "{found_properties[prop][0]}"',
                        ]
                    )
                )
            del found_properties[prop]

    # Issue a warning if there are any unrecognized properties
    if found_properties:
        _logger.warning(
            " ".join(
                [
                    f"Found unrecognized properties in the notation section:",
                    f'{" ".join(found_properties.keys())}',
                ]
            )
        )

    return errors


################################################################################
#                          Exchange section checker                            #
################################################################################
def _verify_bond(lines, line_indices, allowed_atoms) -> bool:
    R"""
    Check that the found "bond" section is following the input file specification.

    Parameters
    ----------
    lines : list of str
        List of the "bond" section lines from the input file.
        Without comments and blank lines.
        ``len(lines) == len(line_indices)``
    line_indices : list of int
        Original line numbers, before filtering.
        ``len(line_indices) == len(lines)``
    allowed_atoms : set of str
        Set of the allowed atom's names.

    Returns
    -------
    errors : bool
        ``True`` if errors are found, ``False`` otherwise.
    """
    # At the beginning we assume that the the bond has at least one line in it
    # with the atom's names and ijk.

    # There are no errors yet
    errors = False

    # We need to make sure that only one entry of the type exists.
    found_data = {"matrix": 0, "symmetric": 0, "dmi": 0, "iso": 0}

    line = lines[0].split()
    # Check that the header line
    # A1 A2 i j k
    if len(line) != 5:
        errors = True
        _logger.error(
            " ".join(
                [
                    f"Line {line_indices[0]}: expected two atom names",
                    f'and three integers separated by spaces, got "{" ".join(line)}"',
                ]
            )
        )
    else:
        # Check the atom names
        errors = errors or _verify_atom_name(line[0], line_indices[0])
        errors = errors or _verify_atom_name(line[1], line_indices[0])
        # Check that atom names are present in the atoms section
        for i in [0, 1]:
            if line[i] not in allowed_atoms:
                errors = True
                _logger.error(
                    " ".join(
                        [
                            f"Line {line_indices[0]}, block {i+1}: atom name {line[i]}",
                            "is not present in the atoms section",
                        ]
                    )
                )

        # Check i j k
        if not (_is_integer(line[2]) and _is_integer(line[3]) and _is_integer(line[4])):
            errors = True
            _logger.error(
                " ".join(
                    [
                        f"Line {line_indices[0]}: expected to have three integers,",
                        f'got "{" ".join(line[2:5])}"',
                    ]
                )
            )

    # Skip first line with atom's names and ijk.
    i = 1
    while i < len(lines):
        line = lines[i].lower().split()
        # If Isotropic keyword found - check isotropic exchange
        # Isotropic Jiso
        if line[0].startswith("i"):
            found_data["iso"] += 1
            # Isotropic line has to have two blocks: keyword and one number
            errors = errors or _verify_numerical_data_block(
                [lines[i]], [[1]], [line_indices[i]], keyword="isotropic"
            )

        # If DMI keyword found - check DMI
        # DMI Dx Dy Dz
        if line[0].startswith("d"):
            found_data["dmi"] += 1
            # DMI line has to have four blocks: keyword and three numbers
            errors = errors or _verify_numerical_data_block(
                [lines[i]], [[3]], [line_indices[i]], keyword="dmi"
            )

        # If symmetric-anisotropy keyword found - check it
        # Symmetric-anisotropy Sxx Syy Sxy Sxz Syz
        if line[0].startswith("s"):
            found_data["symmetric"] += 1
            # Symmetric-anisotropy line has to have six blocks: keyword and five numbers
            errors = errors or _verify_numerical_data_block(
                [lines[i]], [[5]], [line_indices[i]], keyword="symmetric-anisotropy"
            )

        # If Matrix keyword found - check matrix
        # Matrix
        # Jxx Jxy Jxz
        # Jyx Jyy Jyz
        # Jzx Jzy Jzz
        elif line[0].startswith("m"):
            found_data["matrix"] += 1
            errors = errors or _verify_numerical_data_block(
                lines[i : i + 4],
                [0, 3, 3, 3],
                line_indices[i : i + 4],
                keyword="matrix",
            )

        i += 1

    # Check that every type of value was found only once
    total_found_data = 0
    for key in found_data:
        total_found_data += found_data[key]
        if found_data[key] > 1:
            errors = True
            _logger.error(
                " ".join(
                    [
                        f'Line {line_indices[0]}: found more than one "{key}" entry.',
                        f"Check the bond on lines {line_indices[0]}-{line_indices[-1]}",
                    ]
                )
            )

    # Check that at least some values were found
    if total_found_data == 0:
        errors = True
        _logger.error(
            " ".join(
                [
                    f"Line {line_indices[0]}: did not find any information about the parameter value.",
                    f"Check the bond on lines {line_indices[0]}-{line_indices[-1]}",
                ]
            )
        )

    return errors


def _verify_exchange(lines, line_indices, allowed_atoms) -> bool:
    R"""
    Check that the found "exchange" section is following the input file specification.

    Parameters
    ----------
    lines : list of str
        List of the "exchange" section lines from the input file.
        Without comments and blank lines.
        ``len(lines) == len(line_indices)``
    line_indices : list of int
        Original line numbers, before filtering.
        ``len(line_indices) == len(lines)``
    allowed_atoms : set of str
        Set of the allowed atom's names.

    Returns
    -------
    errors : bool
        ``True`` if errors are found, ``False`` otherwise.
    """
    # At the beginning we assume that the the exchange section has at least one line in it
    # with the section header

    # There are no errors yet
    errors = False

    # Check the section header
    errors = errors or _verify_section_header(
        lines[0], line_indices[0], ["m", "e", "j", "k", "r"]
    )

    # Find all bonds.
    # Skip first line with the section header
    i = 1
    found_bonds = []
    while i < len(lines):
        # Skip all possible subsection separators or potential short separators
        while i < len(lines):
            short_separator_errors = _verify_short_separator(
                lines[i], line_indices[i], "-"
            )
            errors = errors or short_separator_errors
            if not (lines[i].startswith("-" * 10) or short_separator_errors):
                break
            i += 1

        # Check that some data is present
        if i >= len(lines):
            break

        bond_start = i
        while i < len(lines) and not lines[i].startswith("-" * 10):
            # Check if the separator is present, but too short
            errors = errors or _verify_short_separator(lines[i], line_indices[i], "-")

            i += 1
        bond_end = i

        found_bonds.append((bond_start, bond_end))

    # exchange section has to have at least one bond
    if len(found_bonds) == 0:
        errors = True
        _logger.error('Found 0 bonds in the "exchange" section, expected at least one')

    for bond in found_bonds:
        errors = errors or _verify_bond(
            lines[slice(*bond)], line_indices[slice(*bond)], allowed_atoms=allowed_atoms
        )

    return errors


################################################################################
#                           On-site section checker                            #
################################################################################
def _verify_on_site(lines, line_indices, allowed_atoms) -> bool:
    r"""
    Check that the found "on-site" section is following the input file specification.

    Parameters
    ----------
    lines : list of str
        List of the "on-site" section lines from the input file.
        Without comments and blank lines.
        ``len(lines) == len(line_indices)``
    line_indices : list of int
        Original line numbers, before filtering.
        ``len(line_indices) == len(lines)``
    allowed_atoms : set of str
        Set of the allowed atom's names.

    Returns
    -------
    errors : bool
        ``True`` if errors are found, ``False`` otherwise.
    """
    # At the beginning we assume that the the on-site section has at least one line in it
    # with the section header

    # There are no errors yet
    errors = False

    # Check the section header
    errors = errors or _verify_section_header(
        lines[0], line_indices[0], ["m", "e", "j", "k", "r"]
    )

    # Find all bonds.
    # Skip first line with the section header
    i = 1
    found_parameters = 0
    while i < len(lines):
        # Skip all possible subsection separators or potential short separators
        while i < len(lines):
            short_separator_errors = _verify_short_separator(
                lines[i], line_indices[i], "-"
            )
            errors = errors or short_separator_errors
            if not (lines[i].startswith("-" * 10) or short_separator_errors):
                break
            i += 1

        # Check that some data is present
        if i >= len(lines):
            break

        # Check that the atom's name is present
        if len(lines[i].split()) != 1:
            errors = True
            _logger.error(
                " ".join(
                    [
                        f"Line {line_indices[i]}: expected only the atom's name,",
                        f'got "{lines[i]}"',
                    ]
                )
            )
        # Check the atom's name
        errors = errors or _verify_atom_name(lines[i].split()[0], line_indices[i])
        # Check that atom names are present in the atoms section
        if lines[i].split()[0] not in allowed_atoms:
            errors = True
            _logger.error(
                " ".join(
                    [
                        f"Line {line_indices[i]}: atom name {lines[i].split()[0]}",
                        "is not present in the atoms section",
                    ]
                )
            )

        # go to the next line
        i += 1
        # Check that the next line is present
        if i >= len(lines):
            errors = True
            _logger.error(
                " ".join(
                    [
                        f"Line {line_indices[i-1]}: Expected to have six numbers,",
                        "separated with spaces at the next line, got nothing",
                    ]
                )
            )
            break

        # Check that the next line contain six numbers
        errors = errors or _verify_numerical_data_block(
            [lines[i]], [6], [line_indices[i]]
        )

        # Go to the next line after verifying the matrix elements
        i += 1
        found_parameters += 1

    # on-site section has to have at least one bond
    if found_parameters == 0:
        errors = True
        _logger.error(
            'Found 0 parameters in the "on-site" section, expected at least one'
        )

    return errors


################################################################################
#                   Rules (check additional ones at the end)                   #
################################################################################

_REQUIRED_SECTIONS = ["cell", "atoms", "notation"]

_SUPPORTED_SECTIONS = {"cell", "atoms", "notation", "exchange", "on-site"}
################################################################################
#                      Mapping of verification functions                       #
################################################################################

_VERIFY = {
    "cell": _verify_cell,
    "atoms": _verify_atoms,
    "notation": _verify_notation,
    "exchange": _verify_exchange,
    "on-site": _verify_on_site,
}


################################################################################
#                              Full file checker                               #
################################################################################
def verify_model_file(lines, line_indices, raise_on_fail=True, return_sections=False):
    r"""
    Verify the content of the input file with the model.

    The input file shall be filtered. See :py:func:`._filter_txt_file`.

    Parameters
    -------------
    lines : list of str
        List of the lines from the input file. Without comments and blank lines.
    line_indices : list of int
        Original line numbers, before filtering.
    raise_on_fail : bool, default True
        Whether to raise an Error if the file content is incorrect.
    return_sections : bool, default False
        Whether to return a dictionary with the positions of the found sections::

            {"keyword" : (start, end)}

        ``lines[start]`` is a first line of the section,
        ``lines[end-1]`` is the last line of the section.
    """

    # There are no errors yet
    errors = False

    # Tracker the found sections
    found_sections = {}

    # Start fir the first line
    i = 0
    while i < len(lines):
        # Skip all possible separators or potential short separators
        while i < len(lines):
            short_separator_errors = _verify_short_separator(
                lines[i], line_indices[i], "="
            )
            errors = errors or short_separator_errors
            if not (lines[i].startswith("=" * 10) or short_separator_errors):
                break
            i += 1

        # Check that there are some data present
        if i >= len(lines):
            break

        section_keyword = lines[i].split()[0].lower()
        section_start = i
        # Iterate until the next section separator is found
        # or the end of the file is reached
        while i < len(lines) and not lines[i].startswith("=" * 10):
            # Check if the separator is present, but too short
            errors = errors or _verify_short_separator(lines[i], line_indices[i], "=")

            i += 1
        section_end = i

        _logger.info(
            f'Found section "{section_keyword}" on lines '
            + f"{line_indices[section_start]}-{line_indices[section_end-1]}"
        )

        # Save the position of the found section
        found_sections[section_keyword] = (section_start, section_end)

    # Check if all required sections are found
    for r_section in _REQUIRED_SECTIONS:
        if r_section not in found_sections:
            errors = True
            _logger.error(f'File: failed to find required section "{r_section}"')

    # Verify atoms section in advance, since we need to know a set of allowed atoms.
    if "atoms" in found_sections:
        allowed_atoms, atoms_errors = _verify_atoms(
            lines[slice(*found_sections["atoms"])],
            line_indices[slice(*found_sections["atoms"])],
        )
        errors = errors or atoms_errors
    else:
        allowed_atoms = set()

    # Set up keywords for the custom verification calls
    kwargs = {}
    # Generic keywords
    for section in _SUPPORTED_SECTIONS:
        kwargs[section] = {}

    # Custom keywords for notation section
    kwargs["notation"]["expect_exchange_factor"] = "exchange" in found_sections
    kwargs["notation"]["expect_on_site_factor"] = "on-site" in found_sections
    kwargs["notation"]["expect_double_counting"] = "exchange" in found_sections
    kwargs["notation"]["expect_spin_normalized"] = (
        "exchange" in found_sections or "on-site" in found_sections
    )

    # Custom keywords for exchange section
    kwargs["exchange"]["allowed_atoms"] = allowed_atoms

    # Custom keywords for on-site section
    kwargs["on-site"]["allowed_atoms"] = allowed_atoms

    # Verify other found sections
    for section in found_sections:
        # Skip atoms section, because it was verified before
        if section == "atoms":
            continue
        # Only verify sections that are supported
        if section in _SUPPORTED_SECTIONS:
            # Check if the verification function is implemented
            if section in _VERIFY:
                errors = errors or _VERIFY[section](
                    lines[slice(*found_sections[section])],
                    line_indices[slice(*found_sections[section])],
                    **kwargs[section],
                )
            # If verification function is not implemented
            else:
                _logger.warning(
                    f"Verification function for the section '{section}' is not implemented."
                )
        # If the section is not supported
        else:
            _logger.warning(f"Section '{section}' is not supported.")

    # Finalize the check and raise an error if needed
    if not errors:
        _logger.info("Model file verification finished: PASSED")
    else:
        if raise_on_fail:
            _logger.info("Model file verification finished: FAILED")
            import sys

            sys.tracebacklimit = 0
            raise FailedToVerifyTxtModelFile("Check log files for the details")
        else:
            _logger.info("Model file verification finished: FAILED")

    if return_sections:
        return found_sections



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

from magnopy._spinham._convention import Convention
from magnopy._spinham._hamiltonian import SpinHamiltonian

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def load_grogu(filename) -> SpinHamiltonian:
    r"""
    Load a SpinHamiltonian object from a .txt file produced by GROGU.

    Parameters
    ----------
    filename : str
        Filename to load SpinHamiltonian object from.

    Returns
    -------
    spinham :py:class:`.SpinHamiltonian`
        SpinHamiltonian object loaded from file.
    """

    # Read the content of the file
    with open(filename, "r", encoding="utf-8") as f:
        lines = f.readlines()

    i = 0
    while "Cell Angstrom" not in lines[i]:
        i += 1

    i += 1

    cell = [
        list(map(float, lines[i].split())),
        list(map(float, lines[i + 1].split())),
        list(map(float, lines[i + 2].split())),
    ]
    i += 2

    while "Atoms Angstrom" not in lines[i]:
        i += 1

    i += 2
    line = lines[i]
    atoms = dict(names=[], positions=[], spins=[], g_factors=[])
    index_map = {}
    atom_index = 0
    while len(lines[i].split()) > 0:
        line = lines[i].split()

        atoms["names"].append(line[0])
        atoms["positions"].append(list(map(float, line[1:4])))
        atoms["spins"].append(np.linalg.norm(list(map(float, line[5:8]))))
        atoms["g_factors"].append(2)
        index_map[line[0]] = atom_index
        atom_index += 1

        i += 1

    convention = Convention(
        multiple_counting=True, spin_normalized=True, c22=0.5, c21=1
    )

    # Construct spin Hamiltonian:
    spinham = SpinHamiltonian(convention=convention, cell=cell, atoms=atoms)

    while "Exchange tensor meV" not in lines[i]:
        i += 1

    i += 4
    while len(lines[i].split()) > 0:
        line = lines[i].split()

        alpha = index_map[line[0]]
        beta = index_map[line[1]]

        nu = tuple(list(map(int, line[2:5])))

        i += 2

        parameter = [
            list(map(float, lines[i].split())),
            list(map(float, lines[i + 1].split())),
            list(map(float, lines[i + 2].split())),
        ]

        i += 4

        spinham.add_22(alpha=alpha, beta=beta, nu=nu, parameter=parameter, replace=True)

    while "Intra-atomic anisotropy tensor meV" not in lines[i]:
        i += 1

    i += 2
    while i < len(lines) and len(lines[i].split()) > 0:
        line = lines[i].split()

        alpha = index_map[line[0]]

        i += 2

        parameter = [
            list(map(float, lines[i].split())),
            list(map(float, lines[i + 1].split())),
            list(map(float, lines[i + 2].split())),
        ]

        i += 4

        spinham.add_21(alpha=alpha, parameter=parameter)

    return spinham


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import os
import sys
from argparse import ArgumentParser, RawDescriptionHelpFormatter

import numpy as np

from magnopy._package_info import logo
from magnopy.io._spin_directions import read_spin_directions
from magnopy.scenarios._solve_lswt import solve_lswt


def manager():
    parser = get_parser()

    args = parser.parse_args()

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    if len(args.spin_directions) == 1:
        args.spin_directions = read_spin_directions(filename=args.spin_directions)
    else:
        args.spin_directions = np.array(args.spin_directions)
        args.spin_directions = args.spin_directions.reshape(
            (len(args.spin_directions) // 3, 3)
        )

    if args.spins is not None:
        args.spins = [float(tmp) for tmp in args.spins]

    kpoints = []
    if args.kpoints is not None:
        with open(args.kpoints, "r") as f:
            for line in f:
                # Remove comment lines
                if line.startswith("#"):
                    continue
                # Remove inline comments and leading/trailing whitespaces
                line = line.split("#")[0].strip()
                # Check for empty lines empty lines
                if line:
                    line = line.split()
                    if len(line) != 3:
                        raise ValueError(
                            f"Expected three numbers per line (in line{i}),"
                            f"got: {len(line)}."
                        )

                    kpoints.append(list(map(float, line)))

        args.kpoints = kpoints

    solve_lswt(**vars(args))


def get_parser():
    parser = ArgumentParser(
        description=logo()
        + "\n\nThis script solves the spin Hamiltonian at the level of "
        "Linear Spin Wave Theory (LSWT) and outputs (almost) every possible quantity.",
        formatter_class=RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "-sf",
        "--spinham-filename",
        type=str,
        metavar="filename",
        default=None,
        required=True,
        help="Path to the spin Hamiltonian file, from where the parameters would be read.",
    )
    parser.add_argument(
        "-ss",
        "--spinham-source",
        type=str,
        metavar="name",
        default=None,
        required=True,
        choices=["GROGU", "TB2J"],
        help='Source of the spin Hamiltonian. Either "GROGU" or "TB2J"',
    )
    parser.add_argument(
        "-sd",
        "--spin-directions",
        nargs="*",
        type=str,
        required=True,
        metavar="S1_x S2_y S3_z ...",
        help="To fully define the system for the calculations of magnons one need the "
        "information about the ground state in addition to the parameters of the "
        "Hamiltonian. There are two ways to give this information to magnopy:\n"
        " * Give a path to the file. In the file there should be M lines with three "
        "numbers in each. The order of the lines would match the order of magnetic "
        "atoms in the spin Hamiltonian."
        " * Give a sequence of 3*M numbers directly to this parameter.",
    )
    parser.add_argument(
        "-s",
        "--spins",
        nargs="*",
        type=str,
        metavar="S1 S2 S3 ...",
        help="In the case when the parameters of spin Hamiltonian comes from TB2J, one "
        "might want to change the values of spins to be closer to half-integers. This "
        "option allows that. Order of the M numbers should match the order of magnetic "
        "atoms in the spin Hamiltonian.",
    )
    parser.add_argument(
        "-kp",
        "--k-path",
        default=None,
        metavar="G-X-S|G-Y",
        type=str,
        help="Path of high symmetry k-points for the plots of dispersion and other "
        "quantities.",
    )
    parser.add_argument(
        "-kps",
        "--kpoints",
        type=str,
        default=None,
        help="Alternatively one could provide an explicit list of k-points for calculation. "
        "In that case provide a path to the file, in which each k-point is given in a "
        "separate line with three numbers per line.",
    )
    parser.add_argument(
        "-r",
        "--relative",
        default=False,
        action="store_true",
        help="When an explicit list of k-points is given, this option specify whether "
        "to consider them as relative or absolute coordinates. Absolute by default.",
    )
    parser.add_argument(
        "-mf",
        "--magnetic-field",
        default=None,
        nargs=3,
        type=float,
        help="Vector of external magnetic field, given in the units of Tesla.",
    )
    parser.add_argument(
        "-of",
        "--output-folder",
        type=str,
        default="magnopy-results",
        help="Folder where all output files of magnopy wil be saved.",
    )
    parser.add_argument(
        "-np",
        "--number-processors",
        type=int,
        default=None,
        help="Number of processes for multithreading. Uses all available processors by "
        "default. Pass 1 to run in serial.",
    )

    return parser



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

from magnopy._spinham._validators import _validate_atom_index


@property
def _p1(spinham) -> list:
    r"""
    Parameters of (one spin & one site) term of the Hamiltonian.

    .. math::

        \boldsymbol{J}_1(\boldsymbol{r}_{\alpha})

    of the term

    .. math::

        C_1
        \sum_{\mu, \alpha, i}
        J_1^i(\boldsymbol{r}_{\alpha})
        S_{\mu,\alpha}^i

    Returns
    -------
    parameters : list
        List of parameters. The list has a form of

        .. code-block:: python

            [[alpha, J], ...]

        ``0 <= len(parameters) <= len(spinham.atoms.names)``.

        where ``alpha`` is an index of the atom to which the parameter is assigned and
        ``J``  is a (3, ) :numpy:`ndarray`. The parameters are sorted by the index of an
        atom ``alpha`` in the ascending order.

    See Also
    --------
    add_1
    remove_1
    """

    return spinham._1


def _add_1(spinham, alpha: int, parameter, replace=False) -> None:
    r"""
    Adds a (one spin & one site) parameter to the Hamiltonian.

    Parameters
    ----------
    alpha : int
        Index of an atom, with which the parameter is associated.

        ``0 <= alpha < len(spinham.atoms.names)``.
    parameter : (3, ) |array-like|_
        Value of the parameter (:math:`3\times1` vector).
    replace : bool, default False
        Whether to replace the value of the parameter if an atom already has a
        parameter associated with it.

    Raises
    ------
    ValueError
        If an atom already has a parameter associated with it.

    See Also
    --------
    p1
    remove_1
    """

    _validate_atom_index(index=alpha, atoms=spinham.atoms)
    spinham._reset_internals()

    parameter = np.array(parameter)

    # TD-BINARY_SEARCH
    # Try to find the place for the new one inside the list
    index = 0
    while index < len(spinham._1):
        # If already present in the model
        if spinham._1[index][0] == alpha:
            # Either replace
            if replace:
                spinham._1[index] = [alpha, parameter]
                return
            # Or raise an error
            raise ValueError(
                f"(One spin & one site) parameter is already set "
                f"for atom {alpha} ('{spinham.atoms.names[alpha]}')"
            )

        # If it should be inserted before current element
        if spinham._1[index][0] > alpha:
            spinham._1.insert(index, [alpha, parameter])
            return

        index += 1

    # If it should be inserted at the end or at the beginning of the list
    spinham._1.append([alpha, parameter])


def _remove_1(spinham, alpha: int) -> None:
    r"""
    Removes a (one spin & one site) parameter from the Hamiltonian.

    Parameters
    ----------
    alpha : int
        Index of an atom, with which the parameter is associated.

        ``0 <= alpha < len(spinham.atoms.names)``.

    See Also
    --------
    p1
    add_1
    """

    _validate_atom_index(index=alpha, atoms=spinham.atoms)

    for i in range(len(spinham._1)):
        # As the list is sorted, there is no point in resuming the search
        # when a larger element is found
        if spinham._1[i][0] > alpha:
            return

        if spinham._1[i][0] == alpha:
            del spinham._1[i]
            spinham._reset_internals()
            return



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import os

import numpy as np
import wulfric

from magnopy._energy import Energy
from magnopy._lswt import LSWT
from magnopy._package_info import logo
from magnopy._parallelization import multiprocess_over_k
from magnopy._spinham._hamiltonian import SpinHamiltonian
from magnopy.io._grogu import load_grogu
from magnopy.io._k_resolved import output_k_resolved, plot_k_resolved
from magnopy.io._spin_directions import read_spin_directions
from magnopy.io._tb2j import load_tb2j

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def solve_lswt(
    spinham_filename: str,
    spinham_source: str,
    spin_directions,
    spins=None,
    k_path=None,
    kpoints=None,
    relative=False,
    magnetic_field=None,
    output_folder="magnopy-results",
    number_processors=None,
) -> None:
    r"""
    Solves the spin Hamiltonian at the level of Linear Spin Wave theory.
    Outputs progress in the standart output (``print()``) and saves some data to
    the files on the disk.

    Parameters
    ----------
    spinham_filename : str
        File with the parameters of spin Hamiltonian.
    spinham_source : str
        Source of the parameters of the spin Hamiltonian.
        One of the "TB2J" or "GROGU". Case-insensitive.
    spin_directions : (M, 3) |array-like|_
        Directions of the local quantization axis for each spin. Magnitude of the vector
        is ignored, only the direction is considered.
    spins : (M, ) |array-like|_, optional
        Spin values. Optional if the spin Hamiltonian is loaded from |TB2J|_. Ignored if
        spin Hamiltonian is loaded from GROGU.
    k_path : str, optional
        Specification of the k-path. The format is "G-X-Y|G-Z" For more details
        on the format see documentation of |Wulfric|_. If nothing given, then the
        k-path is computed by |Wulfric|_ automatically based on the lattice type.
        Ignored if ``kpoints`` are given.
    kpoints : (N, 3) |array-like|_, optional
        Explicit list of k-points to be used instead of automatically generated.
    relative : bool, default False
        If ``relative == True``, then ``kpoints`` are interpreted as given relative to
        the reciprocal unit cell. Otherwise it is interpreted as given in absolute
        coordinates.
    magnetic_field : (3, ) |array-like|_
        Vector of external magnetic field, given in Tesla.
    output_folder : str, default "magnopy-results"
        Name for the folder where to save the output files. If the folder does not exist
        then it will be created.
    number_processors : int, optional
        Number of processors to be used in computation. By default magnopy uses all
        available processes. Use ``number_processors=1`` to run in serial mode.

    Raises
    ------
    ValueError
        If ``spinham_source`` is not supported.
    """

    all_good = True

    print(logo())
    print(f"\n{' Start LSWT ':=^90}\n")

    # Create the output directory if it does not exist
    os.makedirs(output_folder, exist_ok=True)

    # Load spin Hamiltonian
    if spinham_source.lower() == "tb2j":
        spinham = load_tb2j(filename=spinham_filename, spins=spins)
    elif spinham_source.lower() == "grogu":
        spinham = load_grogu(filename=spinham_filename)
    else:
        raise ValueError(
            'Supported sources of spin Hamiltonians are "GROGU" and "TB2J", '
            f'got "{spinham_source}".'
        )

    print(f'Source of the parameters is "{spinham_source.upper()}".')
    print(
        f"Loaded parameters of the spin Hamiltonian from the file\n  "
        f"{os.path.abspath(spinham_filename)}."
    )

    if magnetic_field is not None:
        spinham.add_magnetic_field(h=magnetic_field)

    # Treat kpoints
    if kpoints is not None:
        kpoints = np.array(kpoints, dtype=float)
        if relative:
            kpoints = kpoints @ wulfric.cell.get_reciprocal(cell=spinham.cell)
        kp = None
        flat_indices = None
    else:
        kp = wulfric.Kpoints.from_cell(cell=spinham.cell)
        # Set custom k path
        if k_path is not None:
            kp.path = k_path
        print(f"\n{wulfric.cell.get_variation(spinham.cell)} crystal detected.")
        kpoints = kp.points(relative=False)
        flat_indices = kp.flatten_points(relative=False)

        print("\nList of high symmetry k points")

        header_names = ["k_b1", "k_b2", "k_b3", "k_x", "k_y", "k_z"]
        print(
            " ".join(
                [f"{'name':<5}", f"{'label':>10}"]
                + [f"{tmp:>10}" for tmp in header_names]
            )
        )
        for n_i, name in enumerate(kp.hs_names):
            line = [f"{name:<5}", f"{kp.hs_labels[name]:>10}"]
            for comp in kp.hs_coordinates[name]:
                line.append(f"{comp:>10.6f}")
            rcell = wulfric.cell.get_reciprocal(cell=spinham.cell)
            for comp in kp.hs_coordinates[name] @ rcell:
                line.append(f"{comp:>10.6f}")
            print(" ".join(line))

        print(f'\nK path is "{kp.path_string}"')
        print("\nFlat indices and labels of the points in k path are")
        print(
            "  "
            + "\n  ".join(
                [
                    f"{label:<10} {tick:>12.8f}"
                    for (label, tick) in zip(kp.labels, kp.ticks(relative=False))
                ]
            )
        )

    # Treat spin directions
    spin_directions = np.array(spin_directions, dtype=float)
    spin_directions = (
        spin_directions / np.linalg.norm(spin_directions, axis=1)[:, np.newaxis]
    )

    E_0 = Energy(spinham=spinham).E_0(spin_directions=spin_directions)
    lswt = LSWT(spinham=spinham, spin_directions=spin_directions)

    print(
        f"\n{'Classic ground state energy (E_0)':<51} : "
        f"{E_0:>15.6f} meV\n"
        f"{'Correction to the classic ground state energy (E_2)':<50} : "
        f"{lswt.E_2:>15.6f} meV\n"
    )

    print(
        "Coefficient before one-operator terms (shall be zero if the ground state is correct)"
    )

    print("  " + "\n  ".join([f"{o:12.8f}" for o in lswt.O]))

    if not np.allclose(lswt.O, np.zeros(lswt.O.shape)):
        all_good = False
        print(f"\n{'  WARNING  ':!^90}")
        print(
            "Coefficients before the one-operator terms are not zero. It might indicate  that\n"
            "the ground state (spin directions) is not a ground state of the considered spin\n"
            "Hamiltonian. The results might not be meaningful."
        )
        print(f"{'  END OF WARNING  ':!^90}\n")

    print(f"\nStart calculations over k ... ", end="")
    # Shape (N, M)
    kwargs = dict(relative=relative)

    results = multiprocess_over_k(
        kpoints,
        function=lswt.diagonalize,
        relative=False,
        number_processors=number_processors,
    )

    omegas = np.array([i[0] for i in results])
    deltas = np.array([i[1] for i in results])
    transformation_matrices = np.array([i[2] for i in results])
    print("Done")

    filename = os.path.join(output_folder, "OMEGAS.txt")
    output_k_resolved(
        data=omegas.real,
        data_headers=[f"mode {i+1}" for i in range(len(omegas[0]))],
        output_filename=filename,
        kpoints=kpoints,
        relative=False,
        rcell=wulfric.cell.get_reciprocal(cell=spinham.cell),
        flat_indices=flat_indices,
        digits=6,
        scientific_notation=True,
    )
    print(f"\nOmegas are saved in file\n  {os.path.abspath(filename)}")

    filename = filename[:-4] + ".png"
    plot_k_resolved(
        data=omegas.real,
        kp=kp,
        output_filename=filename,
        ylabel=R"$\omega_{\alpha}(\boldsymbol{k})$, meV",
    )
    print(f"Plot is saved in file\n  {os.path.abspath(filename)}")

    if not np.allclose(omegas.imag, np.zeros(omegas.imag.shape)):
        all_good = False
        print(f"\n{'  WARNING  ':!^90}")
        print(
            "Eigenfrequiencies has non-zero imaginary component for some k vectors. It might\n"
            "indicate that the ground state (spin directions) is not a ground state of the\n"
            "considered spin Hamiltonian. The results might not be meaningful.\n"
        )
        filename = os.path.join(output_folder, "OMEGAS-IMAG.txt")
        output_k_resolved(
            data=omegas.imag,
            data_headers=[f"mode {i+1}" for i in range(len(omegas[0]))],
            output_filename=filename,
            kpoints=kpoints,
            relative=False,
            rcell=wulfric.cell.get_reciprocal(cell=spinham.cell),
            flat_indices=(flat_indices),
            digits=6,
            scientific_notation=True,
        )
        print(
            "Imaginary part of omegas is saved in file\n  "
            f"{os.path.abspath(filename)}"
        )

        filename = filename[:-4] + ".png"
        plot_k_resolved(
            data=omegas.imag,
            kp=kp,
            output_filename=filename,
            ylabel=R"$\mathcal{Im}(\omega_{\alpha}(\boldsymbol{k}))$, meV",
        )
        print(f"Plot of imaginary part is saved in file\n  {os.path.abspath(filename)}")
        print(f"{'  END OF WARNING  ':!^90}\n")

    filename = os.path.join(output_folder, "DELTAS.txt")
    output_k_resolved(
        data=deltas.real,
        data_headers=["Delta"],
        output_filename=filename,
        kpoints=kpoints,
        relative=False,
        rcell=wulfric.cell.get_reciprocal(cell=spinham.cell),
        flat_indices=flat_indices,
        digits=6,
        scientific_notation=True,
    )
    print(f"Deltas are saved in file\n  {os.path.abspath(filename)}")

    filename = filename[:-4] + ".png"
    plot_k_resolved(
        data=deltas.real,
        kp=kp,
        output_filename=filename,
        ylabel=R"$\Delta(\boldsymbol{k})$, meV",
    )
    print(f"Plot is saved in file\n  {os.path.abspath(filename)}")

    if all_good:
        print(f"\n{' Finished OK ':=^90}")
    else:
        print(f"\n{' Finished with WARNINGS ':=^90}")


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from multiprocessing import Pool, cpu_count

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def multiprocess_over_k(kpoints, function, relative=False, number_processors=None):
    r"""
    Parallelize calculation over the kpoints using multiprocessing.

    Parameters
    ----------
    kpoints : (N, 3) |array-like|_
        List of the kpoints.
    function : callable
        Function that prcess one kpoint and will be called as call signature
        ``function(kpoints[i], relative)``.
    relative : bool, default False
        If ``relative=True``, then ``k`` is interpreted as given relative to the
        reciprocal unit cell. Otherwise it is interpreted as given in absolute
        coordinates.
    number_processors : int, optional
        By default use all available processes. Pass ``number_processors=1`` to run in
        serial.

    Returns
    -------
    results : (N, ) list
        List of objects that are returned by the ``function``.
    """

    relative = [relative for _ in kpoints]

    if number_processors == 1:
        results = list(map(function, kpoints, relative))
    else:
        with Pool(number_processors) as p:
            results = p.starmap(function, zip(kpoints, relative))

    return results


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def span_local_rf(direction_vector, hybridize=False):
    r"""
    Span local  right-handed reference frame from the direction vector.

    Parameters
    ----------
    direction_vector : (3, ) |array-like|_
        Direction of the z axis of the local reference frame.
    hybridize : bool, default False
        Whether to return x_alpha, y_alpha, z_alpha or p_alpha, z_alpha.

    Returns
    -------
    x_alpha : (3, ) :numpy:`ndarray`
    y_alpha : (3, ) :numpy:`ndarray`
    p_alpha : (3, ) : numpy:`ndarray`
    z_alpha : (3, ) :numpy:`ndarray`
    """

    direction_vector = np.array(direction_vector, dtype=float)

    if np.allclose(direction_vector, np.zeros(3)):
        raise ValueError("Zero vector.")

    direction_vector /= np.linalg.norm(direction_vector)

    if np.allclose(direction_vector, [0, 0, 1]):
        x_alpha = np.array([1, 0, 0], dtype=float)
        y_alpha = np.array([0, 1, 0], dtype=float)
        z_alpha = np.array([0, 0, 1], dtype=float)
    elif np.allclose(direction_vector, [0, 0, -1]):
        x_alpha = np.array([0, -1, 0], dtype=float)
        y_alpha = np.array([-1, 0, 0], dtype=float)
        z_alpha = np.array([0, 0, -1], dtype=float)
    else:
        z_dir = [0, 0, 1]

        sin_rot_angle = np.linalg.norm(np.cross(z_dir, direction_vector))
        cos_rot_angle = np.dot(z_dir, direction_vector)
        # direction_vector and z_dir are unit vectors
        ux, uy, uz = np.cross(z_dir, direction_vector) / sin_rot_angle

        x_alpha = np.array(
            [
                ux**2 * (1 - cos_rot_angle) + cos_rot_angle,
                ux * uy * (1 - cos_rot_angle) + uz * sin_rot_angle,
                ux * uz * (1 - cos_rot_angle) - uy * sin_rot_angle,
            ]
        )

        y_alpha = np.array(
            [
                ux * uy * (1 - cos_rot_angle) - uz * sin_rot_angle,
                uy**2 * (1 - cos_rot_angle) + cos_rot_angle,
                uy * uz * (1 - cos_rot_angle) + ux * sin_rot_angle,
            ]
        )

    if hybridize:
        return x_alpha + 1j * y_alpha, direction_vector
    return x_alpha, y_alpha, direction_vector


def span_local_rfs(directional_vectors, hybridize=False):
    r"""
    Span local  right-handed reference frames from the direction vectors.

    Parameters
    ----------
    direction_vectors : (M, 3) |array-like|_
        Direction of the z axis of the local reference frames.
    hybridize : bool, default False
        Whether to return x_alpha, y_alpha, z_alpha or p_alpha, z_alpha.

    Returns
    -------
    x_alphas : (M, 3) :numpy:`ndarray`
    y_alphas : (M, 3) :numpy:`ndarray`
    p_alphas : (M, 3) : numpy:`ndarray`
    z_alphas : (M, 3) :numpy:`ndarray`
    """

    results = []

    for directional_vector in directional_vectors:
        results.append(
            span_local_rf(direction_vector=directional_vector, hybridize=hybridize)
        )

    results = np.array(results)

    if hybridize:
        return results[:, 0], results[:, 1]

    return results[:, 0], results[:, 1], results[:, 2]


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging
from argparse import ArgumentParser

import numpy as np

from magnopy._osfix import _winwait
from magnopy._package_info import logo
from magnopy.energy.sub_types import C1, C2, C5
from magnopy.io import load_spinham

_logger = logging.getLogger(__name__)

logging.basicConfig(
    filename="log.txt",
    filemode="a",
    format="%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s",
    datefmt="%H:%M:%S",
    level=logging.DEBUG,
)


def _minimize_C1(energy, spinham, seedname, output_file, spin_output_file):
    output_file.write(
        f"Start to minimize assuming C1 ground state type.\n"
        "Variation with respect to the direction vectors.\n"
    )
    direction_vectors = energy.optimize()
    output_file.write(
        "Minimization for C1 sub-type is done.\n"
        "Spin orientations in the minimum configuration:\n"
    )
    output_file.write(
        f"  {'name':>5} {'r1':>11} {'r2':>11} {'r3':>11} {'sx':>11} {'sy':>11} {'sz':>11}\n"
    )
    for i, spin in enumerate(direction_vectors):
        pos = spinham.magnetic_atoms[i].position
        line = (
            f"  {spinham.magnetic_atoms[i].name:>5} "
            f"{pos[0]:11.8f} {pos[1]:11.8f} {pos[2]:11.8f} "
            f"{spin[0]:11.8f} {spin[1]:11.8f} {spin[2]:11.8f}\n"
        )
        output_file.write(line)
        spin_output_file.write(line)
    direction_vectors /= np.linalg.norm(direction_vectors, axis=1)[:, np.newaxis]
    output_file.write(f"\n Energy: {energy.energy(direction_vectors)}\n")


def _minimize_C2(energy, spinham, seedname, output_file, spin_output_file):
    output_file.write(
        f"Start to minimize assuming C2 ground state type.\n"
        "Variation with respect to the direction vectors, cone axis and spiral vector.\n"
    )
    direction_vectors, cone_axis = energy.optimize()
    output_file.write(
        "Minimization for C2 sub-type is done.\n"
        "Spin orientations in the minimum configuration:\n"
    )
    output_file.write(
        f"  {'name':>5} {'r1':>11} {'r2':>11} {'r3':>11} {'sx':>11} {'sy':>11} {'sz':>11}\n"
    )
    for i, spin in enumerate(direction_vectors):
        pos = spinham.magnetic_atoms[i].position
        line = (
            f"  {spinham.magnetic_atoms[i].name:>5} "
            f"{pos[0]:11.8f} {pos[1]:11.8f} {pos[2]:11.8f} "
            f"{spin[0]:11.8f} {spin[1]:11.8f} {spin[2]:11.8f}\n"
        )
        output_file.write(line)
        spin_output_file.write(line)
    direction_vectors /= np.linalg.norm(direction_vectors, axis=1)[:, np.newaxis]
    output_file.write(
        f"Cone axis: {cone_axis[0]:12.8f} {cone_axis[1]:12.8f} {cone_axis[2]:12.8f}\n"
    )
    output_file.write(f"\n Energy: {energy.energy(direction_vectors, cone_axis)}\n")


def _minimize_C5(energy, spinham, seedname, output_file, spin_output_file):
    output_file.write(
        f"Start to minimize assuming C5 ground state type.\n"
        "Variation with respect to the direction vectors, cone axis and spiral vector.\n"
    )
    direction_vectors, cone_axis, spiral_vector = energy.optimize()
    output_file.write(
        "Minimization for C5 sub-type is done.\n"
        "Spin orientations in the minimum configuration:\n"
    )
    output_file.write(
        f"  {'name':>5} {'r1':>11} {'r2':>11} {'r3':>11} {'sx':>11} {'sy':>11} {'sz':>11}\n"
    )
    for i, spin in enumerate(direction_vectors):
        pos = spinham.magnetic_atoms[i].position
        line = (
            f"  {spinham.magnetic_atoms[i].name:>5} "
            f"{pos[0]:11.8f} {pos[1]:11.8f} {pos[2]:11.8f} "
            f"{spin[0]:11.8f} {spin[1]:11.8f} {spin[2]:11.8f}\n"
        )
        output_file.write(line)
        spin_output_file.write(line)
    direction_vectors /= np.linalg.norm(direction_vectors, axis=1)[:, np.newaxis]
    output_file.write(
        f"Cone axis: {cone_axis[0]:12.8f} {cone_axis[1]:12.8f} {cone_axis[2]:12.8f}\n"
    )
    output_file.write(
        f"Spiral vector: {spiral_vector[0]:12.8f} {spiral_vector[1]:12.8f} {spiral_vector[2]:12.8f}\n"
    )
    output_file.write(
        f"\n Energy: {energy.energy(direction_vectors, cone_axis, spiral_vector)}\n"
    )


def manager(
    spinham,
    spinham_format=None,
    ground_state_type=None,
    magnetic_field=None,
    output_seedname="minim",
):
    # Open main output file
    output_file = open(f"{output_seedname}-results", "w")
    spin_output_file = open(f"{output_seedname}-spins.txt", "w")
    # Write a logo
    output_file.write(logo(date_time=True) + "\n")

    # Load spin Hamiltonian from the input file
    spinham = load_spinham(spinham, spinham_format=spinham_format)

    # Decide which ground state types to minimize
    if "all" in ground_state_type:
        gs_to_minimize = [f"C{i}" for i in range(1, 6)]
    else:
        gs_to_minimize = ground_state_type

    energies = []
    for gs in gs_to_minimize:
        if gs == "C1":
            # Create an instance of the energy class
            energy = C1(spinham)

            # Put some magnetic field into it
            if magnetic_field is not None:
                energy.magnetic_field = magnetic_field
            energies.append(
                _minimize_C1(
                    energy, spinham, output_seedname, output_file, spin_output_file
                )
            )
        elif gs == "C2":
            # Create an instance of the energy class
            energy = C2(spinham)

            # Put some magnetic field into it
            if magnetic_field is not None:
                energy.magnetic_field = magnetic_field
            energies.append(
                _minimize_C2(
                    energy, spinham, output_seedname, output_file, spin_output_file
                )
            )
        elif gs == "C5":
            # Create an instance of the energy class
            energy = C5(spinham)

            # Put some magnetic field into it
            if magnetic_field is not None:
                energy.magnetic_field = magnetic_field
            energies.append(
                _minimize_C5(
                    energy, spinham, output_seedname, output_file, spin_output_file
                )
            )
        else:
            output_file.write(
                f"Minimization for {gs} ground state type is not implemented yet.\n"
            )

    output_file.write("All minimization routines are finished\n")


def create_parser():
    parser = ArgumentParser()
    parser.add_argument(
        "-sh",
        "--spinham",
        type=str,
        required=True,
        help="Relative or absolute path to the file with spin Hamiltonian, "
        "including the name and extension of the file. ",
    )
    parser.add_argument(
        "-shf",
        "--spinham-format",
        type=str,
        choices=["txt", "hdf5", "tb2j"],
        default=None,
        help="Format of the file with spin Hamiltonian.",
    )
    parser.add_argument(
        "-gst",
        "--ground-state-type",
        nargs="*",
        type=str,
        choices=["all", "C1", "C2", "C3", "C4", "C5"],
        default="all",
        help="Type of the ground state to be assumed for the minimization.",
    )
    parser.add_argument(
        "-mf",
        "--magnetic_field",
        default=None,
        type=float,
        nargs=3,
        help="External magnetic field.",
        metavar="Hx Hy Hz",
    )
    parser.add_argument(
        "-os",
        "--output-seedname",
        type=str,
        default="minim",
        help="Seedname for output files",
    )

    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()
    manager(**vars(args))
    _winwait()


if __name__ == "__main__":
    main()



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging

import numpy as np

_logger = logging.getLogger(__name__)

__all__ = []


def _check_directions(n_spins: int, directions):
    r"""
    Check the input correctness for the directional vectors and normalize them.

    Parameters
    ----------
    n_spins : int
        Number of spins in the unit cell (I).
    directions : (I, 3) or (3,) |array-like|_
        Direction of the spin vectors.

    Raises
    ------
    ValueError
        If the input is not correct.

    Notes
    -----
    The vectors are normalized to one, i.e. only the direction of the vectors
    is important.

    Returns
    -------
    directions : (I, 3) :numpy:`ndarray`
        Normalized direction vectors.
    """

    # Check that the input is array-like and convertible to floats
    try:
        directions = np.array(directions, dtype=float)
    except:
        raise ValueError(f"directions is not array-like: {directions}.")

    # Check that the last dimension is 3
    if directions.shape[-1] != 3:
        raise ValueError(
            f"directions has to have last dimension of size 3, "
            f"got {directions.shape[-1]}."
        )

    # Make the input array to have (I,3) shape for I = 1
    if n_spins == 1 and directions.shape == (3,):
        directions = np.array([directions])

    # Check that right amount of direction vectors is given
    if n_spins != directions.shape[0]:
        raise ValueError(
            f"Expected {n_spins} direction vectors, " f"got {directions.shape[0]}."
        )

    # Ensure shape of the form (I,3)
    if directions.shape != (n_spins, 3):
        raise ValueError(
            f"Expected direction vectors to have shape ({n_spins}, 3), "
            f"got {directions.shape}."
        )

    # Normalize direction vectors
    _logger.debug(
        f"Normalized directions, pre-normalized length minimum: "
        f"{np.linalg.norm(directions, axis=1).min():.10f} "
        f"maximum: {np.linalg.norm(directions, axis=1).max():.10f}."
    )
    directions /= np.linalg.norm(directions, axis=1)[..., np.newaxis]

    return directions


def _check_cone_axis(cone_axis):
    r"""
    Check the input correctness for the cone axis and normalize it.

    Parameters
    ----------
    cone_axis : (3,) |array-like|_
        Axis of the spiral cone rotation.

    Raises
    ------
    ValueError
        If the input is not correct.

    Returns
    -------
    cone_axis : (3,) :numpy:`ndarray`
        Normalized cone axis.
    """

    # Check that the input is array-like and convertible to floats
    try:
        cone_axis = np.array(cone_axis, dtype=float)
    except:
        raise ValueError(f"cone_axis is not array-like: {cone_axis}.")

    # Check that the shape and size are correct
    if cone_axis.shape != (3,):
        raise ValueError(f"cone_axis has to have shape (3,), got {cone_axis.shape}.")

    # Normalize cone axis
    _logger.debug(
        f"Normalize cone axis, pre-normalized length "
        f"maximum: {np.linalg.norm(cone_axis):.10f}."
    )
    cone_axis /= np.linalg.norm(cone_axis)

    return cone_axis


def _check_spiral_vector(spiral_vector):
    r"""
    Check the input correctness for the spiral vector.

    Parameters
    ----------
    spiral_vector : (3,) |array-like|_
        Spiral vector.

    Raises
    ------
    ValueError
        If the input is not correct.

    Returns
    -------
    spiral_vector : (3,) :numpy:`ndarray`
        Spiral vector.
    """

    # Check that the input is array-like and convertible to floats
    try:
        spiral_vector = np.array(spiral_vector, dtype=float)
    except:
        raise ValueError(f"spiral_vector is not array-like: {spiral_vector}.")

    # Check that the shape and size are correct
    if spiral_vector.shape != (3,):
        raise ValueError(
            f"spiral_vector has to have shape (3,), got {spiral_vector.shape}."
        )

    return spiral_vector



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

from magnopy._spinham._validators import (
    _spins_ordered,
    _validate_atom_index,
    _validate_unit_cell_index,
)


def _get_primary_p422(alpha, beta, nu, parameter=None):
    r"""
    Return the primary version of the parameter.

    For the definition of the primary version see
    :ref:`user-guide_theory-behind_multiple-counting`.

    Parameters
    ----------
    alpha : int
        Index of the first atom.
    beta : int
        Index of the second atom.
    nu : tuple of 3 int
        Unit cell for the second atom.
    parameter : (3, 3, 3, 3) :numpy:`ndarray`, optional
        Full matrix of the parameter.

    Returns
    -------
    alpha : int
        Index of the first atom.
    beta : int
        Index of the second atom.
    nu : tuple of 3 int
        Unit cell for the second atom.
    parameter : (3, 3, 3, 3) :numpy:`ndarray`
        Full matrix of the parameter. It is returned only if ``parameter is not None``.
    """

    if _spins_ordered(mu1=(0, 0, 0), alpha1=alpha, mu2=nu, alpha2=beta):
        pass
    else:
        i, j, k = nu
        alpha, beta, nu = beta, alpha, (-i, -j, -k)
        if parameter is not None:
            parameter = np.transpose(parameter, (2, 3, 0, 1))

    if parameter is None:
        return alpha, beta, nu
    return alpha, beta, nu, parameter


class _P422_iterator:
    R"""
    Iterator over the (four spins & two sites (2+2)) parameters of the spin Hamiltonian.
    """

    def __init__(self, spinham) -> None:
        self.container = spinham._422
        self.mc = spinham.convention.multiple_counting
        self.length = len(self.container)
        self.index = 0

    def __next__(self):
        if self.index < self.length:
            self.index += 1
            return self.container[self.index - 1]

        elif self.mc and self.index < 2 * self.length:
            self.index += 1
            alpha, beta, (i, j, k), parameter = self.container[
                self.index - 1 - self.length
            ]
            return [beta, alpha, (-i, -j, -k), np.transpose(parameter, (2, 3, 0, 1))]

        raise StopIteration

    def __len__(self):
        return self.length * (1 + int(self.mc))

    def __iter__(self):
        return self


@property
def _p422(spinham):
    r"""
    Parameters of (four spins & two sites (2+2)) term of the Hamiltonian.

    .. math::

        \boldsymbol{J}_{4,2,2}(\boldsymbol{r}_{\nu,\alpha\beta})

    of the term

    .. math::

        C_{4,2,2}
        \sum_{\substack{\mu, \nu, \alpha, \beta,\\ i, j, u, v}}
        J^{ijuv}_{4,2,2}(\boldsymbol{r}_{\nu,\alpha\beta})
        S_{\mu,\alpha}^i
        S_{\mu,\alpha}^j
        S_{\mu+\nu,\beta}^u
        S_{\mu+\nu,\beta}^v

    Returns
    -------
    parameters : iterator
        List of parameters. The list has a form of

        .. code-block:: python

            [[alpha, beta, nu, J], ...]

        where

        ``alpha`` is an index of the atom located in the (0,0,0) unit cell.

        ``beta`` is an index of the atom located in the  nu unit cell.

        ``nu`` defines the unit cell of the second atom (beta). It is a tuple of 3
        integers.

        ``J`` is a (3, 3, 3, 3) :numpy:`ndarray`.

    See Also
    --------
    add_422
    remove_422
    """

    return _P422_iterator(spinham)


def _add_422(
    spinham, alpha: int, beta: int, nu: tuple, parameter, replace=False
) -> None:
    r"""
    Adds a (four spins & two sites (2+2)) parameter to the Hamiltonian.

    Doubles of the bonds are managed automatically (independently of the convention of the
    Hamiltonian).


    Raises
    ------
    ValueError
        If an atom already has a parameter associated with it.

    Parameters
    ----------
    alpha : int
        Index of an atom from the (0, 0, 0) unit cell.

        ``0 <= alpha < len(spinham.atoms.names)``.
    beta : int
        Index of an atom from the nu unit cell.

        ``0 <= beta < len(spinham.atoms.names)``.
    nu : tuple of 3 int
        Three relative coordinates with respect to the three lattice vectors, that
        specify the unit cell for the second atom.

        .. math::

            \nu
            =
            (x_{\boldsymbol{a}_1}, x_{\boldsymbol{a}_2}, x_{\boldsymbol{a}_3})

    parameter : (3, 3, 3, 3) |array-like|_
        Value of the parameter (:math:`3\times3\times3\times3` matrix).
    replace : bool, default False
        Whether to replace the value of the parameter if the pair of atoms
        ``alpha, beta, nu`` or its double already have a parameter associated
        with it.

    See Also
    --------
    p422
    remove_422

    Notes
    -----
    If ``spinham.convention.multiple_counting`` is ``True``, then this function adds both
    the bond and its double to the Hamiltonian. It will cause an ``ValueError`` to
    add the double of the bond after the bond is added.

    If ``spinham.convention.multiple_counting`` is ``False``, then only the primary
    version of the bond is added to the Hamiltonian.

    For the definition of the primary version see
    :ref:`user-guide_theory-behind_multiple-counting`.
    """

    _validate_atom_index(index=alpha, atoms=spinham.atoms)
    _validate_atom_index(index=beta, atoms=spinham.atoms)
    _validate_unit_cell_index(ijk=nu)
    spinham._reset_internals()

    parameter = np.array(parameter)

    alpha, beta, nu, parameter = _get_primary_p422(
        alpha=alpha, beta=beta, nu=nu, parameter=parameter
    )

    # TD-BINARY_SEARCH

    # Try to find the place for the new one inside the list
    index = 0
    while index < len(spinham._422):
        # If already present in the model
        if spinham._422[index][:3] == [alpha, beta, nu]:
            # Either replace
            if replace:
                spinham._422[index] = [alpha, beta, nu, parameter]
                return
            # Or raise an error
            raise ValueError(
                f"Exchange like parameter is already set for the pair of atoms "
                f"{alpha} and {beta} ({nu}). Or for their double bond."
            )

        # If it should be inserted before current element
        if spinham._422[index][:3] > [alpha, beta, nu]:
            spinham._422.insert(index, [alpha, beta, nu, parameter])
            return

        index += 1

    # If it should be inserted at the end or at the beginning of the list
    spinham._422.append([alpha, beta, nu, parameter])


def _remove_422(spinham, alpha: int, beta: int, nu: tuple) -> None:
    r"""
    Removes a (four spins & two sites (2+2)) parameter from the Hamiltonian.

    Doubles of the bonds are managed automatically (independently of the convention of the
    Hamiltonian).

    Parameters
    ----------
    alpha : int
        Index of an atom from the (0, 0, 0) unit cell.

        ``0 <= alpha < len(spinham.atoms.names)``.
    beta : int
        Index of an atom from the nu unit cell.

        ``0 <= beta < len(spinham.atoms.names)``.
    nu : tuple of 3 int
        Three relative coordinates with respect to the three lattice vectors, that
        specify the unit cell for the second atom.

        .. math::

            \nu
            =
            (x_{\boldsymbol{a}_1}, x_{\boldsymbol{a}_2}, x_{\boldsymbol{a}_3})

    See Also
    --------
    p422
    add_422

    Notes
    -----
    If ``spinham.convention.multiple_counting`` is ``True``, then this function removes
    all versions of the bond from the Hamiltonian.

    If ``spinham.convention.multiple_counting`` is ``False``, then this function removes
    the primary version of the given bond.

    For the definition of the primary version see
    :ref:`user-guide_theory-behind_multiple-counting`.

    For instance, if ``(1, 0, (0, 0, 0))`` is given, then this function attempts to
    remove either both ``(1, 0, (0, 0, 0))`` and ``(0, 1, (0, 0, 0))`` if
    ``spinham.convention.multiple_counting == True`` or the primary version
    ``(0, 1, (0, 0, 0))`` if ``spinham.convention.multiple_counting == False``.
    """

    _validate_atom_index(index=alpha, atoms=spinham.atoms)
    _validate_atom_index(index=beta, atoms=spinham.atoms)
    _validate_unit_cell_index(ijk=nu)

    alpha, beta, nu = _get_primary_p422(alpha=alpha, beta=beta, nu=nu)

    # TD-BINARY_SEARCH

    for index in range(len(spinham._422)):
        # As the list is sorted, there is no point in resuming the search
        # when a larger element is found
        if spinham._422[index][:3] > [alpha, beta, nu]:
            return

        if spinham._422[index][:3] == [alpha, beta, nu]:
            del spinham._422[index]
            spinham._reset_internals()
            return



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def read_spin_directions(filename: str):
    r"""
    Read directions of the spins from the file.

    Parameters
    ----------
    filename : str or (3*M, ) |array-like|_
        File with the spin directions. See notes for the specification of the file
        format.

    Returns
    -------
    spin_directions : (M, ) :numpy:`ndarray`
        If ``spin_directions`` is an |array-like|_, then first three elements are

    Notes
    -----
    The file is expected to contain three numbers per line, here is an example for two
    spins

    .. code-block:: text

        S1_x S1_y S1_z
        S2_x S2_y S2_z

    Only the direction of the spin vector is recognized, the modulus is ignored.
    Comments are allowed at any place of the file and preceded by the symbol "#".
    If the symbol "#" is found, the part of the line after it is ignored. Here
    are examples of valid use of the comments

    .. code-block:: text

        # Spin vectors for the material XX
        S1_x S1_y S1_z # Atom X1
        # This comments is here by some reason
        S2_x S2_y S2_z # Atom X2

    """

    spin_directions = []
    with open(spin_directions, "r") as f:
        for line in f:
            # Remove comment lines
            if line.startswith("#"):
                continue
            # Remove inline comments and leading/trailing whitespaces
            line = line.split("#")[0].strip()
            # Check for empty lines empty lines
            if line:
                line = line.split()
                if len(line) != 3:
                    raise ValueError(
                        f"Expected three numbers per line (in line{i}),"
                        f"got: {len(line)}."
                    )
                for tmp in line:
                    spin_directions.append(float(tmp))

    if len(spin_directions) % 3 != 0:
        raise ValueError(
            f"Length of the spin list should be dividable by three, got: {len(spin_directions)}."
        )

    spin_directions = np.array(spin_directions, dtype=float)
    spin_directions = np.reshape(spin_directions, shape=(len(spin_directions) // 3, 3))
    spin_directions = (
        spin_directions / np.linalg.norm(spin_directions, axis=1)[:, np.newaxis]
    )
    return spin_directions


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# MAGNOPY - Python package for magnons.
# Copyright (C) 2023-2025 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
from wulfric.crystal import get_distance

from magnopy._spinham._convention import Convention
from magnopy._spinham._hamiltonian import SpinHamiltonian
from magnopy._spinham._parameter import get_matrix_parameter

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def load_tb2j(filename, spins=None, g_factors=None, quiet=True) -> SpinHamiltonian:
    r"""
    Read spin Hamiltonian from |TB2J|_ output file.

    Parameters
    ----------
    filename : str
        Path to the |TB2J|_ output file.
    spins : (M, ) iterable of floats, optional
        Spin Values for each atom. In the same order as in TB2J file. TB2J outputs
        magnetic moments and not spin values, therefore the spin values for each atom
        are subject to user definition. If user does not define the spin values, then
        magnopy set spin value as :math:`|\boldsymbol{m} / g_factor|` for each spin.
    g_factors : (M, ) iterable of floats, optional
        g-factors for each atom. In the same order as in TB2J file. TB2J does not
        provide g-factors, therefore g-factor for each atom is subject to user
        definition. If user does not define g-factors, then magnopy sets g-factor of
        each atom to :math:`2`.
    quiet : bool, default True
        Whether to suppress output.

    Returns
    -------
    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian build from |TB2J|_ file. Convention is set to the one of TB2J.

    Notes
    -----
    Distances between atoms are not read from the atoms, but rather computed from the
    unit cell and atom positions. If the distance read from the file is different from
    the computed one and ``quiet=False``, the warning is printed. It is usual for the
    computed and read distances to differ in the last digits, since the unit cell is
    provided with the same precision as the distance in |TB2J|_ output file.
    """

    major_sep = "=" * 90
    minor_sep = "-" * 88
    garbage = str.maketrans(
        {"(": None, ")": None, "[": None, "]": None, ",": None, "'": None}
    )
    # Do not correct spelling, it is taken from TB2J.
    cell_flag = "Cell (Angstrom):"
    atoms_flag = "Atoms:"
    atom_end_flag = "Total"
    exchange_flag = "Exchange:"
    iso_flag = "J_iso:"
    aniso_flag = "J_ani:"
    dmi_flag = "DMI:"

    file = open(filename, "r")
    # model = SpinHamiltonian(convention="TB2J")
    line = True

    # Read everything before exchange
    while line:
        line = file.readline()

        # Read cell
        if line and cell_flag in line:
            a1 = file.readline().split()
            a2 = file.readline().split()
            a3 = file.readline().split()

            cell = np.array(
                [
                    list(map(float, a1)),
                    list(map(float, a2)),
                    list(map(float, a3)),
                ]
            )

        # Read atoms
        if line and atoms_flag in line:
            atoms = dict(names=[], positions=[], magnetic_moments=[], charges=[])
            line = file.readline()
            line = file.readline()
            line = file.readline().split()
            i = 0
            while line and atom_end_flag not in line:
                try:
                    # Slicing is not used intentionally.
                    magmom = tuple(map(float, [line[5], line[6], line[7]]))
                except IndexError:
                    magmom = None
                try:
                    charge = float(line[4])
                except IndexError:
                    charge = None

                atoms["names"].append(line[0])
                atoms["positions"].append(np.array(tuple(map(float, line[1:4]))))
                atoms["magnetic_moments"].append(magmom)
                atoms["charges"].append(charge)

                line = file.readline().split()
                i += 1

        # Check if the exchange section is reached
        if line and exchange_flag in line:
            break

    # Populate g_factors of atoms
    if g_factors is None:
        g_factors = [2.0 for _ in range(len(atoms["names"]))]

    atoms["g_factors"] = g_factors

    # Populate spins of atoms
    if spins is None:
        spins = [
            float(np.linalg.norm(atoms["magnetic_moments"][alpha]))
            / atoms["g_factors"][alpha]
            for alpha in range(len(atoms["names"]))
        ]

    atoms["spins"] = spins

    # Create a spin Hamiltonian
    spinham = SpinHamiltonian(
        cell=cell, atoms=atoms, convention=Convention.get_predefined(name="tb2j")
    )

    # Prepare index mapping for atom names
    index_mapping = {}

    # Names of the atoms are unique in the TB2J files
    for index, name in enumerate(spinham.atoms.names):
        index_mapping[name] = index

    # Read exchange (22) parameters
    while line:
        while line and minor_sep not in line:
            line = file.readline()
        line = file.readline().translate(garbage).split()
        atom1 = index_mapping[line[0]]
        atom2 = index_mapping[line[1]]
        ijk = tuple(map(int, line[2:5]))
        distance = float(line[-1])
        iso = None
        aniso = None
        dmi = None
        while line and minor_sep not in line:
            line = file.readline()

            # Read isotropic exchange
            if line and iso_flag in line:
                iso = float(line.split()[-1])

            # Read anisotropic exchange
            if line and aniso_flag in line:
                aniso = np.array(
                    [
                        list(map(float, file.readline().translate(garbage).split())),
                        list(map(float, file.readline().translate(garbage).split())),
                        list(map(float, file.readline().translate(garbage).split())),
                    ]
                )

            # Read DMI
            if line and dmi_flag in line:
                dmi = tuple(map(float, line.translate(garbage).split()[-3:]))

        # Adding info from the exchange block to the SpinHamiltonian structure
        spinham.add_22(
            alpha=atom1,
            beta=atom2,
            nu=ijk,
            # Avoid passing aniso to the function as then the function make it traceless
            # and symmetric, potentially loosing part of the matrix.
            # Due to the TB2J problem: aniso not always traceless.
            parameter=get_matrix_parameter(iso=iso, dmi=dmi) + aniso,
            replace=True,
        )

        computed_distance = get_distance(spinham.cell, spinham.atoms, atom1, atom2, ijk)
        if abs(computed_distance - distance) > 0.001 and not quiet:
            print(
                f"\nComputed distance is a different from the read one:\n"
                + f"  Computed: {computed_distance:.4f}\n  "
                + f"Read: {distance:.4f}\n"
            )

    return spinham


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir
